import pytest

import gitlab


def test_groups(gl):
    # TODO: This one still needs lots of work
    user = gl.users.create(
        {
            "email": "user@test.com",
            "username": "user",
            "name": "user",
            "password": "E4596f8be406Bc3a14a4ccdb1df80587#!1",
        }
    )
    user2 = gl.users.create(
        {
            "email": "user2@test.com",
            "username": "user2",
            "name": "user2",
            "password": "E4596f8be406Bc3a14a4ccdb1df80587#!#2",
        }
    )
    group1 = gl.groups.create(
        {"name": "gitlab-test-group1", "path": "gitlab-test-group1"}
    )
    group2 = gl.groups.create(
        {"name": "gitlab-test-group2", "path": "gitlab-test-group2"}
    )

    p_id = gl.groups.list(search="gitlab-test-group2")[0].id
    group3 = gl.groups.create(
        {"name": "gitlab-test-group3", "path": "gitlab-test-group3", "parent_id": p_id}
    )
    group4 = gl.groups.create(
        {"name": "gitlab-test-group4", "path": "gitlab-test-group4"}
    )

    assert {group1, group2, group3, group4} <= set(gl.groups.list())
    assert gl.groups.list(search="gitlab-test-group1")[0].id == group1.id
    assert group3.parent_id == p_id
    assert group2.subgroups.list()[0].id == group3.id
    assert group2.descendant_groups.list()[0].id == group3.id

    filtered_groups = gl.groups.list(skip_groups=[group3.id, group4.id])
    assert group3 not in filtered_groups
    assert group4 not in filtered_groups

    filtered_groups = gl.groups.list(skip_groups=[group3.id])
    assert group3 not in filtered_groups
    assert group4 in filtered_groups

    group1.members.create(
        {"access_level": gitlab.const.AccessLevel.OWNER, "user_id": user.id}
    )
    group1.members.create(
        {"access_level": gitlab.const.AccessLevel.GUEST, "user_id": user2.id}
    )
    group2.members.create(
        {"access_level": gitlab.const.AccessLevel.OWNER, "user_id": user2.id}
    )

    group4.share(group1.id, gitlab.const.AccessLevel.DEVELOPER)
    group4.share(group2.id, gitlab.const.AccessLevel.MAINTAINER)
    # Reload group4 to have updated shared_with_groups
    group4 = gl.groups.get(group4.id)
    assert len(group4.shared_with_groups) == 2
    group4.unshare(group1.id)
    # Reload group4 to have updated shared_with_groups
    group4 = gl.groups.get(group4.id)
    assert len(group4.shared_with_groups) == 1

    # User memberships (admin only)
    memberships1 = user.memberships.list()
    assert len(memberships1) == 1

    memberships2 = user2.memberships.list()
    assert len(memberships2) == 2

    membership = memberships1[0]
    assert membership.source_type == "Namespace"
    assert membership.access_level == gitlab.const.AccessLevel.OWNER

    project_memberships = user.memberships.list(type="Project")
    assert len(project_memberships) == 0

    group_memberships = user.memberships.list(type="Namespace")
    assert len(group_memberships) == 1

    with pytest.raises(gitlab.GitlabListError) as e:
        membership = user.memberships.list(type="Invalid")
    assert "type does not have a valid value" in str(e.value)

    with pytest.raises(gitlab.GitlabListError) as e:
        user.memberships.list(sudo=user.name)
    assert "403 Forbidden" in str(e.value)

    # Administrator belongs to the groups
    assert len(group1.members.list()) == 3
    assert len(group2.members.list()) == 2

    # Test `user_ids` array
    result = group1.members.list(user_ids=[user.id, 99999])
    assert len(result) == 1
    assert result[0].id == user.id

    group1.members.delete(user.id)

    assert group1.members_all.list()

    member = group1.members.get(user2.id)
    member.access_level = gitlab.const.AccessLevel.OWNER
    member.save()
    member = group1.members.get(user2.id)
    assert member.access_level == gitlab.const.AccessLevel.OWNER

    gl.auth()
    group2.members.delete(gl.user.id)


def test_group_labels(group):
    group.labels.create({"name": "foo", "description": "bar", "color": "#112233"})
    label = group.labels.get("foo")
    assert label.description == "bar"

    label.description = "baz"
    label.save()
    label = group.labels.get("foo")
    assert label.description == "baz"
    assert label in group.labels.list()

    label.new_name = "Label:that requires:encoding"
    label.save()
    assert label.name == "Label:that requires:encoding"
    label = group.labels.get("Label:that requires:encoding")
    assert label.name == "Label:that requires:encoding"

    label.delete()


@pytest.mark.gitlab_premium
@pytest.mark.xfail(reason="/ldap/groups endpoint not documented")
def test_ldap_groups(gl):
    assert isinstance(gl.ldapgroups.list(), list)


@pytest.mark.gitlab_premium
def test_group_ldap_links(group):
    ldap_cn = "common-name"
    ldap_provider = "ldap-provider"

    ldap_cn_link = group.ldap_group_links.create(
        {"provider": ldap_provider, "group_access": 30, "cn": ldap_cn}
    )
    ldap_filter_link = group.ldap_group_links.create(
        {"provider": ldap_provider, "group_access": 30, "filter": "(cn=Common Name)"}
    )

    ldap_links = group.ldap_group_links.list()

    assert ldap_cn_link.cn == ldap_links[0].cn
    assert ldap_filter_link.filter == ldap_links[1].filter

    with pytest.raises(gitlab.GitlabCreateError):
        # todo - can we configure dummy LDAP in the container?
        group.ldap_sync()

    ldap_filter_link.delete()
    group.ldap_group_links.delete(provider=ldap_provider, cn=ldap_cn)

    with pytest.raises(gitlab.GitlabListError, match="No linked LDAP groups found"):
        group.ldap_group_links.list()


def test_group_notification_settings(group):
    settings = group.notificationsettings.get()
    settings.level = "disabled"
    settings.save()

    settings = group.notificationsettings.get()
    assert settings.level == "disabled"


def test_group_badges(group):
    badge_image = "http://example.com"
    badge_link = "http://example/img.svg"
    badge = group.badges.create({"link_url": badge_link, "image_url": badge_image})
    assert badge in group.badges.list()

    badge.image_url = "http://another.example.com"
    badge.save()

    badge = group.badges.get(badge.id)
    assert badge.image_url == "http://another.example.com"

    badge.delete()


def test_group_milestones(group):
    milestone = group.milestones.create({"title": "groupmilestone1"})
    assert milestone in group.milestones.list()

    milestone.due_date = "2020-01-01T00:00:00Z"
    milestone.save()
    milestone.state_event = "close"
    milestone.save()

    milestone = group.milestones.get(milestone.id)
    assert milestone.state == "closed"
    assert not milestone.issues()
    assert not milestone.merge_requests()


def test_group_custom_attributes(gl, group):
    attrs = group.customattributes.list()
    assert not attrs

    attr = group.customattributes.set("key", "value1")
    assert group in gl.groups.list(custom_attributes={"key": "value1"})
    assert attr.key == "key"
    assert attr.value == "value1"
    assert attr in group.customattributes.list()

    attr = group.customattributes.set("key", "value2")
    attr = group.customattributes.get("key")
    assert attr.value == "value2"
    assert attr in group.customattributes.list()

    attr.delete()


def test_group_subgroups_projects(gl, user):
    # TODO: fixture factories
    group1 = gl.groups.list(search="group1")[0]
    group2 = gl.groups.list(search="group2")[0]

    group3 = gl.groups.create(
        {"name": "subgroup1", "path": "subgroup1", "parent_id": group1.id}
    )
    group4 = gl.groups.create(
        {"name": "subgroup2", "path": "subgroup2", "parent_id": group2.id}
    )

    gr1_project = gl.projects.create({"name": "gr1_project", "namespace_id": group1.id})
    gr2_project = gl.projects.create({"name": "gr2_project", "namespace_id": group3.id})

    assert group3.parent_id == group1.id
    assert group4.parent_id == group2.id
    assert gr1_project.namespace["id"] == group1.id
    assert gr2_project.namespace["parent_id"] == group1.id

    gr1_project.delete()
    gr2_project.delete()
    group3.delete()
    group4.delete()


@pytest.mark.gitlab_premium
def test_group_wiki(group):
    content = "Group Wiki page content"
    wiki = group.wikis.create({"title": "groupwikipage", "content": content})
    assert wiki in group.wikis.list()

    wiki = group.wikis.get(wiki.slug)
    assert wiki.content == content

    wiki.content = "new content"
    wiki.save()

    wiki.delete()


@pytest.mark.gitlab_premium
def test_group_hooks(group):
    hook = group.hooks.create({"url": "http://hook.url"})
    assert hook in group.hooks.list()

    hook.note_events = True
    hook.save()

    hook = group.hooks.get(hook.id)
    assert hook.note_events is True

    hook.delete()


def test_group_transfer(gl, group):
    transfer_group = gl.groups.create(
        {"name": "transfer-test-group", "path": "transfer-test-group"}
    )
    transfer_group = gl.groups.get(transfer_group.id)
    assert transfer_group.parent_id != group.id

    transfer_group.transfer(group.id)

    transferred_group = gl.groups.get(transfer_group.id)
    assert transferred_group.parent_id == group.id

    transfer_group.transfer()

    transferred_group = gl.groups.get(transfer_group.id)
    assert transferred_group.path == transferred_group.full_path


@pytest.mark.gitlab_premium
@pytest.mark.xfail(reason="need to setup an identity provider or it's mock")
def test_group_saml_group_links(group):
    group.saml_group_links.create(
        {"saml_group_name": "saml-group-1", "access_level": 10}
    )


@pytest.mark.gitlab_premium
def test_group_service_account(group):
    service_account = group.service_accounts.create(
        {"name": "gitlab-service-account", "username": "gitlab-service-account"}
    )
    assert service_account.name == "gitlab-service-account"
    assert service_account.username == "gitlab-service-account"



"""
GitLab API: https://docs.gitlab.com/ee/api/project_packages_protection_rules.html
"""

import pytest
import responses

from gitlab.v4.objects import ProjectPackageProtectionRule

protected_package_content = {
    "id": 1,
    "project_id": 7,
    "package_name_pattern": "v*",
    "package_type": "npm",
    "minimum_access_level_for_push": "maintainer",
}


@pytest.fixture
def resp_list_protected_packages():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/projects/1/packages/protection/rules",
            json=[protected_package_content],
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_create_protected_package():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.POST,
            url="http://localhost/api/v4/projects/1/packages/protection/rules",
            json=protected_package_content,
            content_type="application/json",
            status=201,
        )
        yield rsps


@pytest.fixture
def resp_update_protected_package():
    updated_content = protected_package_content.copy()
    updated_content["package_name_pattern"] = "abc*"

    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.PATCH,
            url="http://localhost/api/v4/projects/1/packages/protection/rules/1",
            json=updated_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_delete_protected_package():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.DELETE,
            url="http://localhost/api/v4/projects/1/packages/protection/rules/1",
            status=204,
        )
        yield rsps


def test_list_project_protected_packages(project, resp_list_protected_packages):
    protected_package = project.package_protection_rules.list()[0]
    assert isinstance(protected_package, ProjectPackageProtectionRule)
    assert protected_package.package_type == "npm"


def test_create_project_protected_package(project, resp_create_protected_package):
    protected_package = project.package_protection_rules.create(
        {
            "package_name_pattern": "v*",
            "package_type": "npm",
            "minimum_access_level_for_push": "maintainer",
        }
    )
    assert isinstance(protected_package, ProjectPackageProtectionRule)
    assert protected_package.package_type == "npm"


def test_update_project_protected_package(project, resp_update_protected_package):
    updated = project.package_protection_rules.update(
        1, {"package_name_pattern": "abc*"}
    )
    assert updated["package_name_pattern"] == "abc*"


def test_delete_project_protected_package(project, resp_delete_protected_package):
    project.package_protection_rules.delete(1)



import base64
import os
import sys
import tarfile
import time
import zipfile
from io import BytesIO

import pytest

import gitlab


def test_repository_files(project):
    project.files.create(
        {
            "file_path": "README.md",
            "branch": "main",
            "content": "Initial content",
            "commit_message": "Initial commit",
        }
    )
    readme = project.files.get(file_path="README.md", ref="main")
    readme.content = base64.b64encode(b"Improved README").decode()

    time.sleep(2)
    readme.save(branch="main", commit_message="new commit")
    readme.delete(commit_message="Removing README", branch="main")

    project.files.create(
        {
            "file_path": "README.rst",
            "branch": "main",
            "content": "Initial content",
            "commit_message": "New commit",
        }
    )
    readme = project.files.get(file_path="README.rst", ref="main")
    # The first decode() is the ProjectFile method, the second one is the bytes
    # object method
    assert readme.decode().decode() == "Initial content"

    headers = project.files.head("README.rst", ref="main")
    assert headers["X-Gitlab-File-Path"] == "README.rst"

    blame = project.files.blame(file_path="README.rst", ref="main")
    assert blame

    raw_file = project.files.raw(file_path="README.rst", ref="main")
    assert os.fsdecode(raw_file) == "Initial content"

    raw_file = project.files.raw(file_path="README.rst")
    assert os.fsdecode(raw_file) == "Initial content"


def test_repository_tree(project):
    tree = project.repository_tree()
    assert tree
    assert tree[0]["name"] == "README.rst"

    blob_id = tree[0]["id"]
    blob = project.repository_raw_blob(blob_id)
    assert blob.decode() == "Initial content"

    snapshot = project.snapshot()
    assert isinstance(snapshot, bytes)


def test_repository_archive(project):
    archive = project.repository_archive()
    assert isinstance(archive, bytes)

    archive2 = project.repository_archive("main")
    assert archive == archive2


# NOTE(jlvillal): Support for using tarfile.is_tarfile() on a file or file-like object
# was added in Python 3.9
@pytest.mark.skipif(sys.version_info < (3, 9), reason="requires python3.9 or higher")
@pytest.mark.parametrize(
    "format,assertion",
    [
        ("tbz", tarfile.is_tarfile),
        ("tbz2", tarfile.is_tarfile),
        ("tb2", tarfile.is_tarfile),
        ("bz2", tarfile.is_tarfile),
        ("tar", tarfile.is_tarfile),
        ("tar.gz", tarfile.is_tarfile),
        ("tar.bz2", tarfile.is_tarfile),
        ("zip", zipfile.is_zipfile),
    ],
)
def test_repository_archive_formats(project, format, assertion):
    archive = project.repository_archive(format=format)
    assert assertion(BytesIO(archive))


def test_create_commit(project):
    data = {
        "branch": "main",
        "commit_message": "blah blah blah",
        "actions": [{"action": "create", "file_path": "blah", "content": "blah"}],
    }
    commit = project.commits.create(data)

    assert "@@" in project.commits.list()[0].diff()[0]["diff"]
    assert isinstance(commit.refs(), list)
    assert isinstance(commit.merge_requests(), list)


def test_list_all_commits(project):
    data = {
        "branch": "new-branch",
        "start_branch": "main",
        "commit_message": "New commit on new branch",
        "actions": [
            {"action": "create", "file_path": "new-file", "content": "new content"}
        ],
    }
    commit = project.commits.create(data)

    commits = project.commits.list(all=True)
    assert commit not in commits

    # Listing commits on other branches requires `all` parameter passed to the API
    all_commits = project.commits.list(get_all=True, all=True)
    assert commit in all_commits
    assert len(all_commits) > len(commits)


def test_create_commit_status(project):
    commit = project.commits.list()[0]
    status = commit.statuses.create({"state": "success", "sha": commit.id})
    assert status in commit.statuses.list()


def test_commit_signature(project):
    commit = project.commits.list()[0]

    with pytest.raises(gitlab.GitlabGetError) as e:
        commit.signature()

    assert "404 Signature Not Found" in str(e.value)


def test_commit_comment(project):
    commit = project.commits.list()[0]

    commit.comments.create({"note": "This is a commit comment"})
    assert len(commit.comments.list()) == 1


def test_commit_discussion(project):
    commit = project.commits.list()[0]

    discussion = commit.discussions.create({"body": "Discussion body"})
    assert discussion in commit.discussions.list()

    note = discussion.notes.create({"body": "first note"})
    note_from_get = discussion.notes.get(note.id)
    note_from_get.body = "updated body"
    note_from_get.save()
    discussion = commit.discussions.get(discussion.id)

    note_from_get.delete()


def test_revert_commit(project):
    commit = project.commits.list()[0]
    revert_commit = commit.revert(branch="main")

    expected_message = f'Revert "{commit.message}"\n\nThis reverts commit {commit.id}'
    assert revert_commit["message"] == expected_message

    with pytest.raises(gitlab.GitlabRevertError):
        # Two revert attempts should raise GitlabRevertError
        commit.revert(branch="main")


def test_repository_merge_base(project):
    refs = [commit.id for commit in project.commits.list(all=True)]

    commit = project.repository_merge_base(refs)
    assert commit["id"] in refs

    with pytest.raises(gitlab.GitlabGetError, match="Provide at least 2 refs"):
        commit = project.repository_merge_base(refs[0])



import pytest

from gitlab import exceptions


@pytest.mark.parametrize(
    "kwargs,expected",
    [
        ({"error_message": "foo"}, "foo"),
        ({"error_message": "foo", "response_code": "400"}, "400: foo"),
    ],
)
def test_gitlab_error(kwargs, expected):
    error = exceptions.GitlabError(**kwargs)
    assert str(error) == expected


def test_error_raises_from_http_error():
    """Methods decorated with @on_http_error should raise from GitlabHttpError."""

    class TestError(Exception):
        pass

    @exceptions.on_http_error(TestError)
    def raise_error_from_http_error():
        raise exceptions.GitlabHttpError

    with pytest.raises(TestError) as context:
        raise_error_from_http_error()
    assert isinstance(context.value.__cause__, exceptions.GitlabHttpError)



from typing import Any, cast, Union

from gitlab.base import RESTManager, RESTObject
from gitlab.mixins import ListMixin, RetrieveMixin, SaveMixin, UpdateMixin
from gitlab.types import RequiredOptional

__all__ = [
    "ProjectResourceGroup",
    "ProjectResourceGroupManager",
    "ProjectResourceGroupUpcomingJob",
    "ProjectResourceGroupUpcomingJobManager",
]


class ProjectResourceGroup(SaveMixin, RESTObject):
    _id_attr = "key"

    upcoming_jobs: "ProjectResourceGroupUpcomingJobManager"


class ProjectResourceGroupManager(RetrieveMixin, UpdateMixin, RESTManager):
    _path = "/projects/{project_id}/resource_groups"
    _obj_cls = ProjectResourceGroup
    _from_parent_attrs = {"project_id": "id"}
    _list_filters = (
        "order_by",
        "sort",
        "include_html_description",
    )
    _update_attrs = RequiredOptional(optional=("process_mode",))

    def get(
        self, id: Union[str, int], lazy: bool = False, **kwargs: Any
    ) -> ProjectResourceGroup:
        return cast(ProjectResourceGroup, super().get(id=id, lazy=lazy, **kwargs))


class ProjectResourceGroupUpcomingJob(RESTObject):
    pass


class ProjectResourceGroupUpcomingJobManager(ListMixin, RESTManager):
    _path = "/projects/{project_id}/resource_groups/{resource_group_key}/upcoming_jobs"
    _obj_cls = ProjectResourceGroupUpcomingJob
    _from_parent_attrs = {"project_id": "project_id", "resource_group_key": "key"}



"""
GitLab API: https://docs.gitlab.com/ce/api/invitations.html
"""

import re

import pytest
import responses

from gitlab.exceptions import GitlabInvitationError

create_content = {"email": "email@example.com", "access_level": 30}
success_content = {"status": "success"}
error_content = {
    "status": "error",
    "message": {
        "test@example.com": "Invite email has already been taken",
        "test2@example.com": "User already exists in source",
        "test_username": "Access level is not included in the list",
    },
}
invitations_content = [
    {
        "id": 1,
        "invite_email": "member@example.org",
        "created_at": "2020-10-22T14:13:35Z",
        "access_level": 30,
        "expires_at": "2020-11-22T14:13:35Z",
        "user_name": "Raymond Smith",
        "created_by_name": "Administrator",
    },
]
invitation_content = {
    "expires_at": "2012-10-22T14:13:35Z",
    "access_level": 40,
}


@pytest.fixture
def resp_invitations_list():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url=re.compile(r"http://localhost/api/v4/(groups|projects)/1/invitations"),
            json=invitations_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_invitation_create():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.POST,
            url=re.compile(r"http://localhost/api/v4/(groups|projects)/1/invitations"),
            json=success_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_invitation_create_error():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.POST,
            url=re.compile(r"http://localhost/api/v4/(groups|projects)/1/invitations"),
            json=error_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_invitation_update():
    with responses.RequestsMock() as rsps:
        pattern = re.compile(
            r"http://localhost/api/v4/(groups|projects)/1/invitations/email%40example.com"
        )
        rsps.add(
            method=responses.PUT,
            url=pattern,
            json=invitation_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_invitation_delete():
    with responses.RequestsMock() as rsps:
        pattern = re.compile(
            r"http://localhost/api/v4/(groups|projects)/1/invitations/email%40example.com"
        )
        rsps.add(
            method=responses.DELETE,
            url=pattern,
            status=204,
        )
        yield rsps


def test_list_group_invitations(group, resp_invitations_list):
    invitations = group.invitations.list()
    assert invitations[0].invite_email == "member@example.org"


def test_create_group_invitation(group, resp_invitation_create):
    invitation = group.invitations.create(create_content)
    assert invitation.status == "success"


def test_update_group_invitation(group, resp_invitation_update):
    invitation = group.invitations.get("email@example.com", lazy=True)
    invitation.access_level = 30
    invitation.save()


def test_delete_group_invitation(group, resp_invitation_delete):
    invitation = group.invitations.get("email@example.com", lazy=True)
    invitation.delete()
    group.invitations.delete("email@example.com")


def test_list_project_invitations(project, resp_invitations_list):
    invitations = project.invitations.list()
    assert invitations[0].invite_email == "member@example.org"


def test_create_project_invitation(project, resp_invitation_create):
    invitation = project.invitations.create(create_content)
    assert invitation.status == "success"


def test_update_project_invitation(project, resp_invitation_update):
    invitation = project.invitations.get("email@example.com", lazy=True)
    invitation.access_level = 30
    invitation.save()


def test_delete_project_invitation(project, resp_invitation_delete):
    invitation = project.invitations.get("email@example.com", lazy=True)
    invitation.delete()
    project.invitations.delete("email@example.com")


def test_create_group_invitation_raises(group, resp_invitation_create_error):
    with pytest.raises(GitlabInvitationError, match="User already exists"):
        group.invitations.create(create_content)


def test_create_project_invitation_raises(project, resp_invitation_create_error):
    with pytest.raises(GitlabInvitationError, match="User already exists"):
        project.invitations.create(create_content)



package_name = "hello-world"
package_version = "v1.0.0"
file_name = "hello.tar.gz"
file_content = "package content"


def test_list_project_packages(gitlab_cli, project):
    cmd = ["project-package", "list", "--project-id", project.id]
    ret = gitlab_cli(cmd)

    assert ret.success


def test_list_group_packages(gitlab_cli, group):
    cmd = ["group-package", "list", "--group-id", group.id]
    ret = gitlab_cli(cmd)

    assert ret.success


def test_upload_generic_package(tmp_path, gitlab_cli, project):
    path = tmp_path / file_name
    path.write_text(file_content)

    cmd = [
        "-v",
        "generic-package",
        "upload",
        "--project-id",
        project.id,
        "--package-name",
        package_name,
        "--path",
        path,
        "--package-version",
        package_version,
        "--file-name",
        file_name,
    ]
    ret = gitlab_cli(cmd)

    assert "201 Created" in ret.stdout


def test_download_generic_package(gitlab_cli, project):
    cmd = [
        "generic-package",
        "download",
        "--project-id",
        project.id,
        "--package-name",
        package_name,
        "--package-version",
        package_version,
        "--file-name",
        file_name,
    ]
    ret = gitlab_cli(cmd)

    assert ret.stdout == file_content



"""
GitLab API: https://docs.gitlab.com/ce/api/groups.html
"""

import re

import pytest
import responses

import gitlab
from gitlab.v4.objects import (
    GroupDescendantGroup,
    GroupLDAPGroupLink,
    GroupSAMLGroupLink,
    GroupSubgroup,
)
from gitlab.v4.objects.projects import GroupProject, SharedProject

content = {"name": "name", "id": 1, "path": "path"}
ldap_group_links_content = [
    {
        "cn": None,
        "group_access": 40,
        "provider": "ldapmain",
        "filter": "(memberOf=cn=some_group,ou=groups,ou=fake_ou,dc=sub_dc,dc=example,dc=tld)",
    }
]
saml_group_links_content = [{"name": "saml-group-1", "access_level": 10}]
create_saml_group_link_request_body = {
    "saml_group_name": "saml-group-1",
    "access_level": 10,
}
projects_content = [
    {
        "id": 9,
        "description": "foo",
        "default_branch": "master",
        "name": "Html5 Boilerplate",
        "name_with_namespace": "Experimental / Html5 Boilerplate",
        "path": "html5-boilerplate",
        "path_with_namespace": "h5bp/html5-boilerplate",
        "namespace": {"id": 5, "name": "Experimental", "path": "h5bp", "kind": "group"},
    }
]
subgroup_descgroup_content = [
    {
        "id": 2,
        "name": "Bar Group",
        "path": "foo/bar",
        "description": "A subgroup of Foo Group",
        "visibility": "public",
        "share_with_group_lock": False,
        "require_two_factor_authentication": False,
        "two_factor_grace_period": 48,
        "project_creation_level": "developer",
        "auto_devops_enabled": None,
        "subgroup_creation_level": "owner",
        "emails_disabled": None,
        "mentions_disabled": None,
        "lfs_enabled": True,
        "default_branch_protection": 2,
        "avatar_url": "http://gitlab.example.com/uploads/group/avatar/1/bar.jpg",
        "web_url": "http://gitlab.example.com/groups/foo/bar",
        "request_access_enabled": False,
        "full_name": "Bar Group",
        "full_path": "foo/bar",
        "file_template_project_id": 1,
        "parent_id": 123,
        "created_at": "2020-01-15T12:36:29.590Z",
    },
]
push_rules_content = {
    "id": 2,
    "created_at": "2020-08-17T19:09:19.580Z",
    "commit_message_regex": "[a-zA-Z]",
    "commit_message_negative_regex": "[x+]",
    "branch_name_regex": "[a-z]",
    "deny_delete_tag": True,
    "member_check": True,
    "prevent_secrets": True,
    "author_email_regex": "^[A-Za-z0-9.]+@gitlab.com$",
    "file_name_regex": "(exe)$",
    "max_file_size": 100,
}

service_account_content = {
    "name": "gitlab-service-account",
    "username": "gitlab-service-account",
}


@pytest.fixture
def resp_groups():
    with responses.RequestsMock(assert_all_requests_are_fired=False) as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/groups/1",
            json=content,
            content_type="application/json",
            status=200,
        )
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/groups",
            json=[content],
            content_type="application/json",
            status=200,
        )
        rsps.add(
            method=responses.POST,
            url="http://localhost/api/v4/groups",
            json=content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_list_group_projects():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url=re.compile(r"http://localhost/api/v4/groups/1/projects(/shared)?"),
            json=projects_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_list_subgroups_descendant_groups():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url=re.compile(
                r"http://localhost/api/v4/groups/1/(subgroups|descendant_groups)"
            ),
            json=subgroup_descgroup_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_create_import(accepted_content):
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.POST,
            url="http://localhost/api/v4/groups/import",
            json=accepted_content,
            content_type="application/json",
            status=202,
        )
        yield rsps


@pytest.fixture
def resp_transfer_group():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.POST,
            url="http://localhost/api/v4/groups/1/transfer",
            json=content,
            content_type="application/json",
            status=200,
            match=[
                responses.matchers.json_params_matcher({"group_id": "test-namespace"})
            ],
        )
        yield rsps


@pytest.fixture
def resp_list_push_rules_group():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/groups/1/push_rule",
            json=push_rules_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_create_push_rules_group():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.POST,
            url="http://localhost/api/v4/groups/1/push_rule",
            json=push_rules_content,
            content_type="application/json",
            status=201,
        )
        yield rsps


@pytest.fixture
def resp_update_push_rules_group():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/groups/1/push_rule",
            json=push_rules_content,
            content_type="application/json",
            status=200,
        )
        rsps.add(
            method=responses.PUT,
            url="http://localhost/api/v4/groups/1/push_rule",
            json=push_rules_content,
            content_type="application/json",
            status=201,
        )
        yield rsps


@pytest.fixture
def resp_delete_push_rules_group():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/groups/1/push_rule",
            json=push_rules_content,
            content_type="application/json",
            status=200,
        )
        rsps.add(
            method=responses.DELETE,
            url="http://localhost/api/v4/groups/1/push_rule",
            status=204,
        )
        yield rsps


@pytest.fixture
def resp_list_ldap_group_links():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/groups/1/ldap_group_links",
            json=ldap_group_links_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_list_saml_group_links():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/groups/1/saml_group_links",
            json=saml_group_links_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_get_saml_group_link():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/groups/1/saml_group_links/saml-group-1",
            json=saml_group_links_content[0],
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_create_saml_group_link():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.POST,
            url="http://localhost/api/v4/groups/1/saml_group_links",
            match=[
                responses.matchers.json_params_matcher(
                    create_saml_group_link_request_body
                )
            ],
            json=saml_group_links_content[0],
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_delete_saml_group_link():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.POST,
            url="http://localhost/api/v4/groups/1/saml_group_links",
            match=[
                responses.matchers.json_params_matcher(
                    create_saml_group_link_request_body
                )
            ],
            json=saml_group_links_content[0],
            content_type="application/json",
            status=200,
        )
        rsps.add(
            method=responses.DELETE,
            url="http://localhost/api/v4/groups/1/saml_group_links/saml-group-1",
            status=204,
        )
        yield rsps


@pytest.fixture
def resp_restore_group(created_content):
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.POST,
            url="http://localhost/api/v4/groups/1/restore",
            json=created_content,
            content_type="application/json",
            status=201,
        )
        yield rsps


@pytest.fixture
def resp_create_group_service_account():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.POST,
            url="http://localhost/api/v4/groups/1/service_accounts",
            json=service_account_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


def test_get_group(gl, resp_groups):
    data = gl.groups.get(1)
    assert isinstance(data, gitlab.v4.objects.Group)
    assert data.name == "name"
    assert data.path == "path"
    assert data.id == 1


def test_create_group(gl, resp_groups):
    name, path = "name", "path"
    data = gl.groups.create({"name": name, "path": path})
    assert isinstance(data, gitlab.v4.objects.Group)
    assert data.name == name
    assert data.path == path


def test_create_group_export(group, resp_export):
    export = group.exports.create()
    assert export.message == "202 Accepted"


def test_list_group_projects(group, resp_list_group_projects):
    projects = group.projects.list()
    assert isinstance(projects[0], GroupProject)
    assert projects[0].path == projects_content[0]["path"]


def test_list_group_shared_projects(group, resp_list_group_projects):
    projects = group.shared_projects.list()
    assert isinstance(projects[0], SharedProject)
    assert projects[0].path == projects_content[0]["path"]


def test_list_group_subgroups(group, resp_list_subgroups_descendant_groups):
    subgroups = group.subgroups.list()
    assert isinstance(subgroups[0], GroupSubgroup)
    assert subgroups[0].path == subgroup_descgroup_content[0]["path"]


def test_list_group_descendant_groups(group, resp_list_subgroups_descendant_groups):
    descendant_groups = group.descendant_groups.list()
    assert isinstance(descendant_groups[0], GroupDescendantGroup)
    assert descendant_groups[0].path == subgroup_descgroup_content[0]["path"]


def test_list_ldap_group_links(group, resp_list_ldap_group_links):
    ldap_group_links = group.ldap_group_links.list()
    assert isinstance(ldap_group_links[0], GroupLDAPGroupLink)
    assert ldap_group_links[0].provider == ldap_group_links_content[0]["provider"]


@pytest.mark.skip("GitLab API endpoint not implemented")
def test_refresh_group_export_status(group, resp_export):
    export = group.exports.create()
    export.refresh()
    assert export.export_status == "finished"


def test_download_group_export(group, resp_export, binary_content):
    export = group.exports.create()
    download = export.download()
    assert isinstance(download, bytes)
    assert download == binary_content


def test_import_group(gl, resp_create_import):
    group_import = gl.groups.import_group("file", "api-group", "API Group")
    assert group_import["message"] == "202 Accepted"


@pytest.mark.skip("GitLab API endpoint not implemented")
def test_refresh_group_import_status(group, resp_groups):
    group_import = group.imports.get()
    group_import.refresh()
    assert group_import.import_status == "finished"


def test_transfer_group(gl, resp_transfer_group):
    group = gl.groups.get(1, lazy=True)
    group.transfer("test-namespace")


def test_list_group_push_rules(group, resp_list_push_rules_group):
    pr = group.pushrules.get()
    assert pr
    assert pr.deny_delete_tag


def test_create_group_push_rule(group, resp_create_push_rules_group):
    group.pushrules.create({"deny_delete_tag": True})


def test_update_group_push_rule(
    group,
    resp_update_push_rules_group,
):
    pr = group.pushrules.get()
    pr.deny_delete_tag = False
    pr.save()


def test_delete_group_push_rule(group, resp_delete_push_rules_group):
    pr = group.pushrules.get()
    pr.delete()


def test_list_saml_group_links(group, resp_list_saml_group_links):
    saml_group_links = group.saml_group_links.list()
    assert isinstance(saml_group_links[0], GroupSAMLGroupLink)
    assert saml_group_links[0].name == saml_group_links_content[0]["name"]
    assert (
        saml_group_links[0].access_level == saml_group_links_content[0]["access_level"]
    )


def test_get_saml_group_link(group, resp_get_saml_group_link):
    saml_group_link = group.saml_group_links.get("saml-group-1")
    assert isinstance(saml_group_link, GroupSAMLGroupLink)
    assert saml_group_link.name == saml_group_links_content[0]["name"]
    assert saml_group_link.access_level == saml_group_links_content[0]["access_level"]


def test_create_saml_group_link(group, resp_create_saml_group_link):
    saml_group_link = group.saml_group_links.create(create_saml_group_link_request_body)
    assert isinstance(saml_group_link, GroupSAMLGroupLink)
    assert (
        saml_group_link.name == create_saml_group_link_request_body["saml_group_name"]
    )
    assert (
        saml_group_link.access_level
        == create_saml_group_link_request_body["access_level"]
    )


def test_delete_saml_group_link(group, resp_delete_saml_group_link):
    saml_group_link = group.saml_group_links.create(create_saml_group_link_request_body)
    saml_group_link.delete()


def test_group_restore(group, resp_restore_group):
    group.restore()


def test_create_group_service_account(group, resp_create_group_service_account):
    service_account = group.service_accounts.create(
        {"name": "gitlab-service-account", "username": "gitlab-service-account"}
    )
    assert service_account.name == "gitlab-service-account"
    assert service_account.username == "gitlab-service-account"



import datetime
import io
import json
from typing import Optional

import requests
import responses

from gitlab import base

MATCH_EMPTY_QUERY_PARAMS = [responses.matchers.query_param_matcher({})]


class FakeObject(base.RESTObject):
    pass


class FakeObjectWithoutId(base.RESTObject):
    _id_attr = None


class FakeObjectWithLongRepr(base.RESTObject):
    _id_attr = None
    _repr_attr = "test"


class OtherFakeObject(FakeObject):
    _id_attr = "foo"


class FakeManager(base.RESTManager):
    _path = "/tests"
    _obj_cls = FakeObject


class FakeParent(FakeObject):
    id = 42


class FakeManagerWithParent(base.RESTManager):
    _path = "/tests/{test_id}/cases"
    _obj_cls = FakeObject
    _from_parent_attrs = {"test_id": "id"}


# NOTE: The function `httmock_response` and the class `Headers` is taken from
# https://github.com/patrys/httmock/ which is licensed under the Apache License, Version
# 2.0. Thus it is allowed to be used in this project.
# https://www.apache.org/licenses/GPL-compatibility.html
class Headers(object):
    def __init__(self, res):
        self.headers = res.headers

    def get_all(self, name, failobj=None):
        return self.getheaders(name)

    def getheaders(self, name):
        return [self.headers.get(name)]


def httmock_response(
    status_code: int = 200,
    content: str = "",
    headers=None,
    reason=None,
    elapsed=0,
    request: Optional[requests.models.PreparedRequest] = None,
    stream: bool = False,
    http_vsn=11,
) -> requests.models.Response:
    res = requests.Response()
    res.status_code = status_code
    if isinstance(content, (dict, list)):
        content = json.dumps(content).encode("utf-8")
    if isinstance(content, str):
        content = content.encode("utf-8")
    res._content = content
    res._content_consumed = content
    res.headers = requests.structures.CaseInsensitiveDict(headers or {})
    res.encoding = requests.utils.get_encoding_from_headers(res.headers)
    res.reason = reason
    res.elapsed = datetime.timedelta(elapsed)
    res.request = request
    if hasattr(request, "url"):
        res.url = request.url
        if isinstance(request.url, bytes):
            res.url = request.url.decode("utf-8")
    if "set-cookie" in res.headers:
        res.cookies.extract_cookies(
            requests.cookies.MockResponse(Headers(res)),
            requests.cookies.MockRequest(request),
        )
    if stream:
        res.raw = io.BytesIO(content)
    else:
        res.raw = io.BytesIO(b"")
    res.raw.version = http_vsn

    # normally this closes the underlying connection,
    #  but we have nothing to free.
    res.close = lambda *args, **kwargs: None

    return res



import datetime
import time

import pytest

import gitlab
import gitlab.v4.objects


def test_merge_requests(project):
    project.files.create(
        {
            "file_path": "README.rst",
            "branch": "main",
            "content": "Initial content",
            "commit_message": "Initial commit",
        }
    )

    source_branch = "branch-merge-request-api"
    project.branches.create({"branch": source_branch, "ref": "main"})

    project.files.create(
        {
            "file_path": "README2.rst",
            "branch": source_branch,
            "content": "Initial content",
            "commit_message": "New commit in new branch",
        }
    )
    project.mergerequests.create(
        {"source_branch": source_branch, "target_branch": "main", "title": "MR readme2"}
    )


def test_merge_requests_get(project, merge_request):
    mr = project.mergerequests.get(merge_request.iid)
    assert mr.iid == merge_request.iid

    mr = project.mergerequests.get(str(merge_request.iid))
    assert mr.iid == merge_request.iid


@pytest.mark.gitlab_premium
def test_merge_requests_list_approver_ids(project):
    # show https://github.com/python-gitlab/python-gitlab/issues/1698 is now
    # fixed
    project.mergerequests.list(
        all=True,
        state="opened",
        author_id=423,
        approver_ids=[423],
    )


def test_merge_requests_get_lazy(project, merge_request):
    mr = project.mergerequests.get(merge_request.iid, lazy=True)
    assert mr.iid == merge_request.iid


def test_merge_request_discussion(project):
    mr = project.mergerequests.list()[0]

    discussion = mr.discussions.create({"body": "Discussion body"})
    assert discussion in mr.discussions.list()

    note = discussion.notes.create({"body": "first note"})
    note_from_get = discussion.notes.get(note.id)
    note_from_get.body = "updated body"
    note_from_get.save()

    discussion = mr.discussions.get(discussion.id)
    assert discussion.attributes["notes"][-1]["body"] == "updated body"

    note_from_get.delete()


def test_merge_request_labels(project):
    mr = project.mergerequests.list()[0]
    mr.labels = ["label2"]
    mr.save()

    events = mr.resourcelabelevents.list()
    assert events

    event = mr.resourcelabelevents.get(events[0].id)
    assert event


def test_merge_request_milestone_events(project, milestone):
    mr = project.mergerequests.list()[0]
    mr.milestone_id = milestone.id
    mr.save()

    milestones = mr.resourcemilestoneevents.list()
    assert milestones

    milestone = mr.resourcemilestoneevents.get(milestones[0].id)
    assert milestone


def test_merge_request_basic(project):
    mr = project.mergerequests.list()[0]
    # basic testing: only make sure that the methods exist
    mr.commits()
    mr.changes()
    participants = mr.participants()
    assert participants
    assert isinstance(participants, list)


def test_merge_request_rebase(project):
    mr = project.mergerequests.list()[0]
    assert mr.rebase()


@pytest.mark.gitlab_premium
@pytest.mark.xfail(reason="project /approvers endpoint is gone")
def test_project_approvals(project):
    mr = project.mergerequests.list()[0]
    approval = project.approvals.get()

    reset_value = approval.reset_approvals_on_push
    approval.reset_approvals_on_push = not reset_value
    approval.save()

    approval = project.approvals.get()
    assert reset_value != approval.reset_approvals_on_push

    project.approvals.set_approvers([1], [])
    approval = project.approvals.get()
    assert approval.approvers[0]["user"]["id"] == 1

    approval = mr.approvals.get()
    approval.approvals_required = 2
    approval.save()
    approval = mr.approvals.get()
    assert approval.approvals_required == 2

    approval.approvals_required = 3
    approval.save()
    approval = mr.approvals.get()
    assert approval.approvals_required == 3

    mr.approvals.set_approvers(1, [1], [])
    approval = mr.approvals.get()
    assert approval.approvers[0]["user"]["id"] == 1


@pytest.mark.gitlab_premium
def test_project_merge_request_approval_rules(group, project):
    approval_rules = project.approvalrules.list(get_all=True)
    assert not approval_rules

    project.approvalrules.create(
        {"name": "approval-rule", "approvals_required": 2, "group_ids": [group.id]}
    )
    approval_rules = project.approvalrules.list(get_all=True)
    assert len(approval_rules) == 1
    assert approval_rules[0].approvals_required == 2

    approval_rules[0].save()
    approval_rules = project.approvalrules.list(get_all=True)
    assert len(approval_rules) == 1
    assert approval_rules[0].approvals_required == 2

    approval_rules[0].delete()


def test_merge_request_reset_approvals(gitlab_url, project):
    today = datetime.date.today()
    future_date = today + datetime.timedelta(days=4)
    bot = project.access_tokens.create(
        {"name": "bot", "scopes": ["api"], "expires_at": future_date.isoformat()}
    )

    bot_gitlab = gitlab.Gitlab(gitlab_url, private_token=bot.token)
    bot_project = bot_gitlab.projects.get(project.id, lazy=True)

    # Pause to let GL catch up (happens on hosted too, sometimes takes a while for server to be ready to merge)
    time.sleep(5)

    mr = bot_project.mergerequests.list()[0]  # type: ignore[index]

    assert mr.reset_approvals()


def test_cancel_merge_when_pipeline_succeeds(project, merge_request_with_pipeline):
    # Pause to let GL catch up (happens on hosted too, sometimes takes a while for server to be ready to merge)
    time.sleep(5)
    # Set to merge when the pipeline succeeds, which should never happen
    merge_request_with_pipeline.merge(merge_when_pipeline_succeeds=True)
    # Pause to let GL catch up (happens on hosted too, sometimes takes a while for server to be ready to merge)
    time.sleep(5)

    mr = project.mergerequests.get(merge_request_with_pipeline.iid)
    assert mr.merged_at is None
    assert mr.merge_when_pipeline_succeeds is True
    cancel = mr.cancel_merge_when_pipeline_succeeds()
    assert cancel == {"status": "success"}


def test_merge_request_merge(project, merge_request):
    merge_request.merge()
    # Pause to let GL catch up (happens on hosted too, sometimes takes a while for server to be ready to merge)
    time.sleep(5)

    mr = project.mergerequests.get(merge_request.iid)
    assert mr.merged_at is not None
    assert mr.merge_when_pipeline_succeeds is False
    with pytest.raises(gitlab.GitlabMRClosedError):
        # Two merge attempts should raise GitlabMRClosedError
        mr.merge()


def test_merge_request_should_remove_source_branch(project, merge_request) -> None:
    """Test to ensure
    https://github.com/python-gitlab/python-gitlab/issues/1120 is fixed.
    Bug reported that they could not use 'should_remove_source_branch' in
    mr.merge() call"""
    merge_request.merge(should_remove_source_branch=True)
    # Pause to let GL catch up (happens on hosted too, sometimes takes a while for server to be ready to merge)
    time.sleep(5)

    # Wait until it is merged
    mr = None
    mr_iid = merge_request.iid
    for _ in range(60):
        mr = project.mergerequests.get(mr_iid)
        if mr.merged_at is not None:
            break
        time.sleep(0.5)

    assert mr is not None
    assert mr.merged_at is not None
    time.sleep(0.5)
    # Pause to let GL catch up (happens on hosted too, sometimes takes a while for server to be ready to merge)
    time.sleep(5)

    # Ensure we can NOT get the MR branch
    with pytest.raises(gitlab.exceptions.GitlabGetError):
        result = project.branches.get(merge_request.source_branch)
        # Help to debug in case the expected exception doesn't happen.
        import pprint

        print("mr:", pprint.pformat(mr))
        print("mr.merged_at:", pprint.pformat(mr.merged_at))
        print("result:", pprint.pformat(result))


def test_merge_request_large_commit_message(project, merge_request) -> None:
    """Test to ensure https://github.com/python-gitlab/python-gitlab/issues/1452
    is fixed.
    Bug reported that very long 'merge_commit_message' in mr.merge() would
    cause an error: 414 Request too large
    """
    merge_commit_message = "large_message\r\n" * 1_000
    assert len(merge_commit_message) > 10_000

    merge_request.merge(
        merge_commit_message=merge_commit_message, should_remove_source_branch=False
    )

    # Pause to let GL catch up (happens on hosted too, sometimes takes a while for server to be ready to merge)
    time.sleep(5)

    # Wait until it is merged
    mr = None
    mr_iid = merge_request.iid
    for _ in range(60):
        mr = project.mergerequests.get(mr_iid)
        if mr.merged_at is not None:
            break
        time.sleep(0.5)

    assert mr is not None
    assert mr.merged_at is not None
    time.sleep(0.5)

    # Ensure we can get the MR branch
    project.branches.get(merge_request.source_branch)


def test_merge_request_merge_ref(merge_request) -> None:
    response = merge_request.merge_ref()
    assert response and "commit_id" in response


def test_merge_request_merge_ref_should_fail(project, merge_request) -> None:
    # Create conflict
    project.files.create(
        {
            "file_path": f"README.{merge_request.source_branch}",
            "branch": project.default_branch,
            "content": "Different initial content",
            "commit_message": "Another commit in main branch",
        }
    )
    # Pause to let GL catch up (happens on hosted too, sometimes takes a while for server to be ready to merge)
    time.sleep(5)

    # Check for non-existing merge_ref for MR with conflicts
    with pytest.raises(gitlab.exceptions.GitlabGetError):
        response = merge_request.merge_ref()
        assert "commit_id" not in response



"""
GitLab API: https://docs.gitlab.com/ee/api/statistics.html
"""


def test_get_statistics(gl):
    statistics = gl.statistics.get()

    assert statistics.snippets.isdigit()
    assert statistics.users.isdigit()
    assert statistics.groups.isdigit()
    assert statistics.projects.isdigit()



"""
GitLab API:
https://docs.gitlab.com/ee/api/releases/index.html
https://docs.gitlab.com/ee/api/releases/links.html
"""

import re

import pytest
import responses

from gitlab.v4.objects import ProjectReleaseLink

tag_name = "v1.0.0"
release_name = "demo-release"
release_description = "my-rel-desc"
released_at = "2019-03-15T08:00:00Z"
link_name = "hello-world"
link_url = "https://gitlab.example.com/group/hello/-/jobs/688/artifacts/raw/bin/hello-darwin-amd64"
direct_url = f"https://gitlab.example.com/group/hello/-/releases/{tag_name}/downloads/hello-world"
new_link_type = "package"
link_content = {
    "id": 2,
    "name": link_name,
    "url": link_url,
    "direct_asset_url": direct_url,
    "external": False,
    "link_type": "other",
}

release_content = {
    "id": 3,
    "tag_name": tag_name,
    "name": release_name,
    "description": release_description,
    "milestones": [],
    "released_at": released_at,
}

release_url = re.compile(rf"http://localhost/api/v4/projects/1/releases/{tag_name}")
links_url = re.compile(
    rf"http://localhost/api/v4/projects/1/releases/{tag_name}/assets/links"
)
link_id_url = re.compile(
    rf"http://localhost/api/v4/projects/1/releases/{tag_name}/assets/links/1"
)


@pytest.fixture
def resp_list_links():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url=links_url,
            json=[link_content],
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_get_link():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url=link_id_url,
            json=link_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_create_link():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.POST,
            url=links_url,
            json=link_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_update_link():
    updated_content = dict(link_content)
    updated_content["link_type"] = new_link_type

    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.PUT,
            url=link_id_url,
            json=updated_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_delete_link():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.DELETE,
            url=link_id_url,
            status=204,
        )
        yield rsps


@pytest.fixture
def resp_update_release():
    updated_content = dict(release_content)

    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.PUT,
            url=release_url,
            json=updated_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


def test_list_release_links(release, resp_list_links):
    links = release.links.list()
    assert isinstance(links, list)
    assert isinstance(links[0], ProjectReleaseLink)
    assert links[0].url == link_url


def test_get_release_link(release, resp_get_link):
    link = release.links.get(1)
    assert isinstance(link, ProjectReleaseLink)
    assert link.url == link_url


def test_create_release_link(release, resp_create_link):
    link = release.links.create({"url": link_url, "name": link_name})
    assert isinstance(link, ProjectReleaseLink)
    assert link.url == link_url


def test_update_release_link(release, resp_update_link):
    link = release.links.get(1, lazy=True)
    link.link_type = new_link_type
    link.save()
    assert link.link_type == new_link_type


def test_delete_release_link(release, resp_delete_link):
    link = release.links.get(1, lazy=True)
    link.delete()


def test_update_release(release, resp_update_release):
    release.name = release_name
    release.description = release_description
    release.save()
    assert release.name == release_name
    assert release.description == release_description



from gitlab.base import RESTManager, RESTObject
from gitlab.mixins import CreateMixin
from gitlab.types import RequiredOptional

__all__ = ["GroupServiceAccount", "GroupServiceAccountManager"]


class GroupServiceAccount(RESTObject):
    pass


class GroupServiceAccountManager(CreateMixin, RESTManager):
    _path = "/groups/{group_id}/service_accounts"
    _obj_cls = GroupServiceAccount
    _from_parent_attrs = {"group_id": "id"}
    _create_attrs = RequiredOptional(
        optional=("name", "username"),
    )



import pytest

from gitlab import types


class TestRequiredOptional:
    def test_requiredoptional_empty(self):
        b = types.RequiredOptional()
        assert not b.required
        assert not b.optional
        assert not b.exclusive

    def test_requiredoptional_values_no_keywords(self):
        b = types.RequiredOptional(
            ("required1", "required2"),
            ("optional1", "optional2"),
            ("exclusive1", "exclusive2"),
        )
        assert b.required == ("required1", "required2")
        assert b.optional == ("optional1", "optional2")
        assert b.exclusive == ("exclusive1", "exclusive2")

    def test_requiredoptional_values_keywords(self):
        b = types.RequiredOptional(
            exclusive=("exclusive1", "exclusive2"),
            optional=("optional1", "optional2"),
            required=("required1", "required2"),
        )
        assert b.required == ("required1", "required2")
        assert b.optional == ("optional1", "optional2")
        assert b.exclusive == ("exclusive1", "exclusive2")

    def test_validate_attrs_required(self):
        data = {"required1": 1, "optional2": 2}
        rq = types.RequiredOptional(required=("required1",))
        rq.validate_attrs(data=data)
        data = {"optional1": 1, "optional2": 2}
        with pytest.raises(AttributeError, match="Missing attributes: required1"):
            rq.validate_attrs(data=data)

    def test_validate_attrs_exclusive(self):
        data = {"exclusive1": 1, "optional1": 1}
        rq = types.RequiredOptional(exclusive=("exclusive1", "exclusive2"))
        rq.validate_attrs(data=data)
        data = {"exclusive1": 1, "exclusive2": 2, "optional1": 1}
        with pytest.raises(
            AttributeError,
            match="Provide only one of these attributes: exclusive1, exclusive2",
        ):
            rq.validate_attrs(data=data)


def test_gitlab_attribute_get():
    o = types.GitlabAttribute("whatever")
    assert o.get() == "whatever"

    o.set_from_cli("whatever2")
    assert o.get() == "whatever2"
    assert o.get_for_api(key="spam") == ("spam", "whatever2")

    o = types.GitlabAttribute()
    assert o._value is None


def test_array_attribute_input():
    o = types.ArrayAttribute()
    o.set_from_cli("foo,bar,baz")
    assert o.get() == ["foo", "bar", "baz"]

    o.set_from_cli("foo")
    assert o.get() == ["foo"]


def test_array_attribute_empty_input():
    o = types.ArrayAttribute()
    o.set_from_cli("")
    assert o.get() == []

    o.set_from_cli("  ")
    assert o.get() == []


def test_array_attribute_get_for_api_from_cli():
    o = types.ArrayAttribute()
    o.set_from_cli("foo,bar,baz")
    assert o.get_for_api(key="spam") == ("spam[]", ["foo", "bar", "baz"])


def test_array_attribute_get_for_api_from_list():
    o = types.ArrayAttribute(["foo", "bar", "baz"])
    assert o.get_for_api(key="spam") == ("spam[]", ["foo", "bar", "baz"])


def test_array_attribute_get_for_api_from_int_list():
    o = types.ArrayAttribute([1, 9, 7])
    assert o.get_for_api(key="spam") == ("spam[]", [1, 9, 7])


def test_array_attribute_does_not_split_string():
    o = types.ArrayAttribute("foo")
    assert o.get_for_api(key="spam") == ("spam[]", "foo")


# CommaSeparatedListAttribute tests
def test_csv_string_attribute_get_for_api_from_cli():
    o = types.CommaSeparatedListAttribute()
    o.set_from_cli("foo,bar,baz")
    assert o.get_for_api(key="spam") == ("spam", "foo,bar,baz")


def test_csv_string_attribute_get_for_api_from_list():
    o = types.CommaSeparatedListAttribute(["foo", "bar", "baz"])
    assert o.get_for_api(key="spam") == ("spam", "foo,bar,baz")


def test_csv_string_attribute_get_for_api_from_int_list():
    o = types.CommaSeparatedListAttribute([1, 9, 7])
    assert o.get_for_api(key="spam") == ("spam", "1,9,7")


# LowercaseStringAttribute tests
def test_lowercase_string_attribute_get_for_api():
    o = types.LowercaseStringAttribute("FOO")
    assert o.get_for_api(key="spam") == ("spam", "foo")



from typing import Any, cast

from gitlab.base import RESTManager, RESTObject
from gitlab.mixins import CreateMixin, DownloadMixin, GetWithoutIdMixin, RefreshMixin
from gitlab.types import RequiredOptional

__all__ = [
    "GroupExport",
    "GroupExportManager",
    "GroupImport",
    "GroupImportManager",
    "ProjectExport",
    "ProjectExportManager",
    "ProjectImport",
    "ProjectImportManager",
]


class GroupExport(DownloadMixin, RESTObject):
    _id_attr = None


class GroupExportManager(GetWithoutIdMixin, CreateMixin, RESTManager):
    _path = "/groups/{group_id}/export"
    _obj_cls = GroupExport
    _from_parent_attrs = {"group_id": "id"}

    def get(self, **kwargs: Any) -> GroupExport:
        return cast(GroupExport, super().get(**kwargs))


class GroupImport(RESTObject):
    _id_attr = None


class GroupImportManager(GetWithoutIdMixin, RESTManager):
    _path = "/groups/{group_id}/import"
    _obj_cls = GroupImport
    _from_parent_attrs = {"group_id": "id"}

    def get(self, **kwargs: Any) -> GroupImport:
        return cast(GroupImport, super().get(**kwargs))


class ProjectExport(DownloadMixin, RefreshMixin, RESTObject):
    _id_attr = None


class ProjectExportManager(GetWithoutIdMixin, CreateMixin, RESTManager):
    _path = "/projects/{project_id}/export"
    _obj_cls = ProjectExport
    _from_parent_attrs = {"project_id": "id"}
    _create_attrs = RequiredOptional(optional=("description",))

    def get(self, **kwargs: Any) -> ProjectExport:
        return cast(ProjectExport, super().get(**kwargs))


class ProjectImport(RefreshMixin, RESTObject):
    _id_attr = None


class ProjectImportManager(GetWithoutIdMixin, RESTManager):
    _path = "/projects/{project_id}/import"
    _obj_cls = ProjectImport
    _from_parent_attrs = {"project_id": "id"}

    def get(self, **kwargs: Any) -> ProjectImport:
        return cast(ProjectImport, super().get(**kwargs))



import pytest
import requests

import gitlab


@pytest.fixture(
    scope="session",
    params=[{"get_all": True}, {"all": True}],
    ids=["get_all=True", "all=True"],
)
def get_all_kwargs(request):
    """A tiny parametrized fixture to inject both `get_all=True` and
    `all=True` to ensure they behave the same way for pagination."""
    return request.param


def test_auth_from_config(gl, gitlab_config, temp_dir):
    """Test token authentication from config file"""
    test_gitlab = gitlab.Gitlab.from_config(config_files=[gitlab_config])
    test_gitlab.auth()
    assert isinstance(test_gitlab.user, gitlab.v4.objects.CurrentUser)


def test_no_custom_session(gl, temp_dir):
    """Test no custom session"""
    custom_session = requests.Session()
    test_gitlab = gitlab.Gitlab.from_config(
        config_files=[temp_dir / "python-gitlab.cfg"]
    )
    assert test_gitlab.session != custom_session


def test_custom_session(gl, temp_dir):
    """Test custom session"""
    custom_session = requests.Session()
    test_gitlab = gitlab.Gitlab.from_config(
        config_files=[temp_dir / "python-gitlab.cfg"], session=custom_session
    )
    assert test_gitlab.session == custom_session


def test_broadcast_messages(gl, get_all_kwargs):
    msg = gl.broadcastmessages.create({"message": "this is the message"})
    msg.color = "#444444"
    msg.save()
    msg_id = msg.id

    msg = gl.broadcastmessages.list(**get_all_kwargs)[0]
    assert msg.color == "#444444"

    msg = gl.broadcastmessages.get(msg_id)
    assert msg.color == "#444444"

    msg.delete()


def test_markdown(gl):
    html = gl.markdown("foo")
    assert "foo" in html


def test_markdown_in_project(gl, project):
    html = gl.markdown("foo", project=project.path_with_namespace)
    assert "foo" in html


def test_sidekiq_queue_metrics(gl):
    out = gl.sidekiq.queue_metrics()
    assert isinstance(out, dict)
    assert "default" in out["queues"]


def test_sidekiq_process_metrics(gl):
    out = gl.sidekiq.process_metrics()
    assert isinstance(out, dict)
    assert "hostname" in out["processes"][0]


def test_sidekiq_job_stats(gl):
    out = gl.sidekiq.job_stats()
    assert isinstance(out, dict)
    assert "processed" in out["jobs"]


def test_sidekiq_compound_metrics(gl):
    out = gl.sidekiq.compound_metrics()
    assert isinstance(out, dict)
    assert "jobs" in out
    assert "processes" in out
    assert "queues" in out


@pytest.mark.gitlab_premium
def test_geo_nodes(gl):
    # Very basic geo nodes tests because we only have 1 node.
    nodes = gl.geonodes.list()
    assert isinstance(nodes, list)

    status = gl.geonodes.status()
    assert isinstance(status, list)


@pytest.mark.gitlab_premium
def test_gitlab_license(gl):
    license = gl.get_license()
    assert "user_limit" in license

    with pytest.raises(gitlab.GitlabLicenseError, match="The license key is invalid."):
        gl.set_license("dummy key")


def test_gitlab_settings(gl):
    settings = gl.settings.get()
    settings.default_projects_limit = 42
    settings.save()
    settings = gl.settings.get()
    assert settings.default_projects_limit == 42


def test_template_dockerfile(gl):
    assert gl.dockerfiles.list()

    dockerfile = gl.dockerfiles.get("Node")
    assert dockerfile.content is not None


def test_template_gitignore(gl, get_all_kwargs):
    assert gl.gitignores.list(**get_all_kwargs)
    gitignore = gl.gitignores.get("Node")
    assert gitignore.content is not None


def test_template_gitlabciyml(gl, get_all_kwargs):
    assert gl.gitlabciymls.list(**get_all_kwargs)
    gitlabciyml = gl.gitlabciymls.get("Nodejs")
    assert gitlabciyml.content is not None


def test_template_license(gl):
    assert gl.licenses.list()
    license = gl.licenses.get(
        "bsd-2-clause", project="mytestproject", fullname="mytestfullname"
    )
    assert "mytestfullname" in license.content


def test_hooks(gl):
    hook = gl.hooks.create({"url": "http://whatever.com"})
    assert hook in gl.hooks.list()

    hook.delete()


def test_namespaces(gl, get_all_kwargs):
    gl.auth()
    current_user = gl.user.username

    namespaces = gl.namespaces.list(**get_all_kwargs)
    assert namespaces

    namespaces = gl.namespaces.list(search=current_user, **get_all_kwargs)
    assert namespaces[0].kind == "user"

    namespace = gl.namespaces.get(current_user)
    assert namespace.kind == "user"

    namespace = gl.namespaces.exists(current_user)
    assert namespace.exists


def test_notification_settings(gl):
    settings = gl.notificationsettings.get()
    settings.level = gitlab.const.NotificationLevel.WATCH
    settings.save()

    settings = gl.notificationsettings.get()
    assert settings.level == gitlab.const.NotificationLevel.WATCH


def test_search(gl):
    result = gl.search(scope=gitlab.const.SearchScope.USERS, search="Administrator")
    assert result[0]["id"] == 1


def test_user_activities(gl):
    activities = gl.user_activities.list(query_parameters={"from": "2019-01-01"})
    assert isinstance(activities, list)


def test_events(gl):
    events = gl.events.list()
    assert isinstance(events, list)


@pytest.mark.skip
def test_features(gl):
    feat = gl.features.set("foo", 30)
    assert feat.name == "foo"
    assert feat in gl.features.list()

    feat.delete()


def test_pagination(gl, project):
    project2 = gl.projects.create({"name": "project-page-2"})

    list1 = gl.projects.list(per_page=1, page=1)
    list2 = gl.projects.list(per_page=1, page=2)
    assert len(list1) == 1
    assert len(list2) == 1
    assert list1[0].id != list2[0].id

    project2.delete()


def test_rate_limits(gl):
    settings = gl.settings.get()
    settings.throttle_authenticated_api_enabled = True
    settings.throttle_authenticated_api_requests_per_period = 1
    settings.throttle_authenticated_api_period_in_seconds = 3
    settings.save()

    projects = []
    for i in range(0, 20):
        projects.append(gl.projects.create({"name": f"{str(i)}ok"}))

    with pytest.raises(gitlab.GitlabCreateError) as e:
        for i in range(20, 40):
            projects.append(
                gl.projects.create(
                    {"name": f"{str(i)}shouldfail"}, obey_rate_limit=False
                )
            )

    assert "Retry later" in str(e.value)

    settings.throttle_authenticated_api_enabled = False
    settings.save()
    [project.delete() for project in projects]


def test_list_default_warning(gl):
    """When there are more than 20 items and use default `list()` then warning is
    generated"""
    with pytest.warns(UserWarning, match="python-gitlab.readthedocs.io") as record:
        gl.gitlabciymls.list()

    assert len(record) == 1
    warning = record[0]
    assert __file__ == warning.filename
    assert __file__ in str(warning.message)


def test_list_page_nowarning(gl, recwarn):
    """Using `page=X` will disable the warning"""
    gl.gitlabciymls.list(page=1)
    assert not recwarn


def test_list_all_false_nowarning(gl, recwarn):
    """Using `all=False` will disable the warning"""
    gl.gitlabciymls.list(all=False)
    assert not recwarn


def test_list_all_true_nowarning(gl, get_all_kwargs, recwarn):
    """Using `get_all=True` will disable the warning"""
    items = gl.gitlabciymls.list(**get_all_kwargs)
    for warn in recwarn:
        if issubclass(warn.category, UserWarning):
            # Our warning has a link to the docs in it, make sure we don't have
            # that.
            assert "python-gitlab.readthedocs.io" not in str(warn.message)
    assert len(items) > 20


def test_list_iterator_true_nowarning(gl, recwarn):
    """Using `iterator=True` will disable the warning"""
    items = gl.gitlabciymls.list(iterator=True)
    assert not recwarn
    assert len(list(items)) > 20



from typing import Any, cast, Union

from gitlab.base import RESTManager, RESTObject
from gitlab.mixins import (
    CreateMixin,
    DeleteMixin,
    ObjectDeleteMixin,
    ObjectRotateMixin,
    RetrieveMixin,
    RotateMixin,
)
from gitlab.types import ArrayAttribute, RequiredOptional

__all__ = [
    "ProjectAccessToken",
    "ProjectAccessTokenManager",
]


class ProjectAccessToken(ObjectDeleteMixin, ObjectRotateMixin, RESTObject):
    pass


class ProjectAccessTokenManager(
    CreateMixin, DeleteMixin, RetrieveMixin, RotateMixin, RESTManager
):
    _path = "/projects/{project_id}/access_tokens"
    _obj_cls = ProjectAccessToken
    _from_parent_attrs = {"project_id": "id"}
    _create_attrs = RequiredOptional(
        required=("name", "scopes"), optional=("access_level", "expires_at")
    )
    _types = {"scopes": ArrayAttribute}

    def get(
        self, id: Union[str, int], lazy: bool = False, **kwargs: Any
    ) -> ProjectAccessToken:
        return cast(ProjectAccessToken, super().get(id=id, lazy=lazy, **kwargs))



import copy
import warnings

import pytest
import requests
import responses

from gitlab import GitlabHttpError, GitlabList, GitlabParsingError, RedirectError
from gitlab.const import RETRYABLE_TRANSIENT_ERROR_CODES
from tests.unit import helpers


def test_build_url(gl):
    r = gl._build_url("http://localhost/api/v4")
    assert r == "http://localhost/api/v4"
    r = gl._build_url("https://localhost/api/v4")
    assert r == "https://localhost/api/v4"
    r = gl._build_url("/projects")
    assert r == "http://localhost/api/v4/projects"


@responses.activate
def test_http_request(gl):
    url = "http://localhost/api/v4/projects"
    responses.add(
        method=responses.GET,
        url=url,
        json=[{"name": "project1"}],
        status=200,
        match=helpers.MATCH_EMPTY_QUERY_PARAMS,
    )

    http_r = gl.http_request("get", "/projects")
    http_r.json()
    assert http_r.status_code == 200
    assert responses.assert_call_count(url, 1) is True


@responses.activate
def test_http_request_with_url_encoded_kwargs_does_not_duplicate_params(gl):
    url = "http://localhost/api/v4/projects?topics%5B%5D=python"
    responses.add(
        method=responses.GET,
        url=url,
        json=[{"name": "project1"}],
        status=200,
        match=[responses.matchers.query_param_matcher({"topics[]": "python"})],
    )

    kwargs = {"topics[]": "python"}
    http_r = gl.http_request("get", "/projects?topics%5B%5D=python", **kwargs)
    http_r.json()
    assert http_r.status_code == 200
    assert responses.assert_call_count(url, 1)


@responses.activate
def test_http_request_404(gl):
    url = "http://localhost/api/v4/not_there"
    responses.add(
        method=responses.GET,
        url=url,
        json={},
        status=400,
        match=helpers.MATCH_EMPTY_QUERY_PARAMS,
    )

    with pytest.raises(GitlabHttpError):
        gl.http_request("get", "/not_there")
    assert responses.assert_call_count(url, 1) is True


@responses.activate
@pytest.mark.parametrize("status_code", RETRYABLE_TRANSIENT_ERROR_CODES)
def test_http_request_with_only_failures(gl, status_code):
    url = "http://localhost/api/v4/projects"
    responses.add(
        method=responses.GET,
        url=url,
        json={},
        status=status_code,
        match=helpers.MATCH_EMPTY_QUERY_PARAMS,
    )

    with pytest.raises(GitlabHttpError):
        gl.http_request("get", "/projects")

    assert responses.assert_call_count(url, 1) is True


@responses.activate
def test_http_request_with_retry_on_method_for_transient_failures(gl):
    call_count = 0
    calls_before_success = 3

    url = "http://localhost/api/v4/projects"

    def request_callback(request):
        nonlocal call_count
        call_count += 1
        status_code = 200 if call_count >= calls_before_success else 500
        headers = {}
        body = "[]"

        return (status_code, headers, body)

    responses.add_callback(
        method=responses.GET,
        url=url,
        callback=request_callback,
        content_type="application/json",
    )

    http_r = gl.http_request("get", "/projects", retry_transient_errors=True)

    assert http_r.status_code == 200
    assert len(responses.calls) == calls_before_success


@responses.activate
@pytest.mark.parametrize(
    "exception",
    [
        requests.ConnectionError("Connection aborted."),
        requests.exceptions.ChunkedEncodingError("Connection broken."),
    ],
)
def test_http_request_with_retry_on_method_for_transient_network_failures(
    gl, exception
):
    call_count = 0
    calls_before_success = 3

    url = "http://localhost/api/v4/projects"

    def request_callback(request):
        nonlocal call_count
        call_count += 1
        status_code = 200
        headers = {}
        body = "[]"

        if call_count >= calls_before_success:
            return (status_code, headers, body)
        raise exception

    responses.add_callback(
        method=responses.GET,
        url=url,
        callback=request_callback,
        content_type="application/json",
    )

    http_r = gl.http_request("get", "/projects", retry_transient_errors=True)

    assert http_r.status_code == 200
    assert len(responses.calls) == calls_before_success


@responses.activate
def test_http_request_with_retry_on_class_for_transient_failures(gl_retry):
    call_count = 0
    calls_before_success = 3

    url = "http://localhost/api/v4/projects"

    def request_callback(request: requests.models.PreparedRequest):
        nonlocal call_count
        call_count += 1
        status_code = 200 if call_count >= calls_before_success else 500
        headers = {}
        body = "[]"

        return (status_code, headers, body)

    responses.add_callback(
        method=responses.GET,
        url=url,
        callback=request_callback,
        content_type="application/json",
    )

    http_r = gl_retry.http_request("get", "/projects", retry_transient_errors=True)

    assert http_r.status_code == 200
    assert len(responses.calls) == calls_before_success


@responses.activate
def test_http_request_with_retry_on_class_for_transient_network_failures(gl_retry):
    call_count = 0
    calls_before_success = 3

    url = "http://localhost/api/v4/projects"

    def request_callback(request: requests.models.PreparedRequest):
        nonlocal call_count
        call_count += 1
        status_code = 200
        headers = {}
        body = "[]"

        if call_count >= calls_before_success:
            return (status_code, headers, body)
        raise requests.ConnectionError("Connection aborted.")

    responses.add_callback(
        method=responses.GET,
        url=url,
        callback=request_callback,
        content_type="application/json",
    )

    http_r = gl_retry.http_request("get", "/projects", retry_transient_errors=True)

    assert http_r.status_code == 200
    assert len(responses.calls) == calls_before_success


@responses.activate
def test_http_request_with_retry_on_class_and_method_for_transient_failures(gl_retry):
    call_count = 0
    calls_before_success = 3

    url = "http://localhost/api/v4/projects"

    def request_callback(request):
        nonlocal call_count
        call_count += 1
        status_code = 200 if call_count >= calls_before_success else 500
        headers = {}
        body = "[]"

        return (status_code, headers, body)

    responses.add_callback(
        method=responses.GET,
        url=url,
        callback=request_callback,
        content_type="application/json",
    )

    with pytest.raises(GitlabHttpError):
        gl_retry.http_request("get", "/projects", retry_transient_errors=False)

    assert len(responses.calls) == 1


@responses.activate
def test_http_request_with_retry_on_class_and_method_for_transient_network_failures(
    gl_retry,
):
    call_count = 0
    calls_before_success = 3

    url = "http://localhost/api/v4/projects"

    def request_callback(request):
        nonlocal call_count
        call_count += 1
        status_code = 200
        headers = {}
        body = "[]"

        if call_count >= calls_before_success:
            return (status_code, headers, body)
        raise requests.ConnectionError("Connection aborted.")

    responses.add_callback(
        method=responses.GET,
        url=url,
        callback=request_callback,
        content_type="application/json",
    )

    with pytest.raises(requests.ConnectionError):
        gl_retry.http_request("get", "/projects", retry_transient_errors=False)

    assert len(responses.calls) == 1


def create_redirect_response(
    *, response: requests.models.Response, http_method: str, api_path: str
) -> requests.models.Response:
    """Create a Requests response object that has a redirect in it"""

    assert api_path.startswith("/")
    http_method = http_method.upper()

    # Create a history which contains our original request which is redirected
    history = [
        helpers.httmock_response(
            status_code=302,
            content="",
            headers={"Location": f"http://example.com/api/v4{api_path}"},
            reason="Moved Temporarily",
            request=response.request,
        )
    ]

    # Create a "prepped" Request object to be the final redirect. The redirect
    # will be a "GET" method as Requests changes the method to "GET" when there
    # is a 301/302 redirect code.
    req = requests.Request(
        method="GET",
        url=f"http://example.com/api/v4{api_path}",
    )
    prepped = req.prepare()

    resp_obj = helpers.httmock_response(
        status_code=200,
        content="",
        headers={},
        reason="OK",
        elapsed=5,
        request=prepped,
    )
    resp_obj.history = history
    return resp_obj


def test_http_request_302_get_does_not_raise(gl):
    """Test to show that a redirect of a GET will not cause an error"""

    method = "get"
    api_path = "/user/status"
    url = f"http://localhost/api/v4{api_path}"

    def response_callback(
        response: requests.models.Response,
    ) -> requests.models.Response:
        return create_redirect_response(
            response=response, http_method=method, api_path=api_path
        )

    with responses.RequestsMock(response_callback=response_callback) as req_mock:
        req_mock.add(
            method=responses.GET,
            url=url,
            status=302,
            match=helpers.MATCH_EMPTY_QUERY_PARAMS,
        )
        gl.http_request(verb=method, path=api_path)


def test_http_request_302_put_raises_redirect_error(gl):
    """Test to show that a redirect of a PUT will cause an error"""

    method = "put"
    api_path = "/user/status"
    url = f"http://localhost/api/v4{api_path}"

    def response_callback(
        response: requests.models.Response,
    ) -> requests.models.Response:
        return create_redirect_response(
            response=response, http_method=method, api_path=api_path
        )

    with responses.RequestsMock(response_callback=response_callback) as req_mock:
        req_mock.add(
            method=responses.PUT,
            url=url,
            status=302,
            match=helpers.MATCH_EMPTY_QUERY_PARAMS,
        )
        with pytest.raises(RedirectError) as exc:
            gl.http_request(verb=method, path=api_path)
    error_message = exc.value.error_message
    assert "Moved Temporarily" in error_message
    assert "http://localhost/api/v4/user/status" in error_message
    assert "http://example.com/api/v4/user/status" in error_message


def test_http_request_on_409_resource_lock_retries(gl_retry):
    url = "http://localhost/api/v4/user"
    retried = False

    def response_callback(
        response: requests.models.Response,
    ) -> requests.models.Response:
        """We need a callback that adds a resource lock reason only on first call"""
        nonlocal retried

        if not retried:
            response.reason = "Resource lock"

        retried = True
        return response

    with responses.RequestsMock(response_callback=response_callback) as rsps:
        rsps.add(
            method=responses.GET,
            url=url,
            status=409,
            match=helpers.MATCH_EMPTY_QUERY_PARAMS,
        )
        rsps.add(
            method=responses.GET,
            url=url,
            status=200,
            match=helpers.MATCH_EMPTY_QUERY_PARAMS,
        )
        response = gl_retry.http_request("get", "/user")

    assert response.status_code == 200


def test_http_request_on_409_resource_lock_without_retry_raises(gl):
    url = "http://localhost/api/v4/user"

    def response_callback(
        response: requests.models.Response,
    ) -> requests.models.Response:
        """Without retry, this will fail on the first call"""
        response.reason = "Resource lock"
        return response

    with responses.RequestsMock(response_callback=response_callback) as req_mock:
        req_mock.add(
            method=responses.GET,
            url=url,
            status=409,
            match=helpers.MATCH_EMPTY_QUERY_PARAMS,
        )
        with pytest.raises(GitlabHttpError) as excinfo:
            gl.http_request("get", "/user")

    assert excinfo.value.response_code == 409


@responses.activate
def test_get_request(gl):
    url = "http://localhost/api/v4/projects"
    responses.add(
        method=responses.GET,
        url=url,
        json={"name": "project1"},
        status=200,
        match=helpers.MATCH_EMPTY_QUERY_PARAMS,
    )

    result = gl.http_get("/projects")
    assert isinstance(result, dict)
    assert result["name"] == "project1"
    assert responses.assert_call_count(url, 1) is True


@responses.activate
def test_get_request_raw(gl):
    url = "http://localhost/api/v4/projects"
    responses.add(
        method=responses.GET,
        url=url,
        content_type="application/octet-stream",
        body="content",
        status=200,
        match=helpers.MATCH_EMPTY_QUERY_PARAMS,
    )

    result = gl.http_get("/projects")
    assert result.content.decode("utf-8") == "content"
    assert responses.assert_call_count(url, 1) is True


@responses.activate
def test_get_request_404(gl):
    url = "http://localhost/api/v4/not_there"
    responses.add(
        method=responses.GET,
        url=url,
        json=[],
        status=404,
        match=helpers.MATCH_EMPTY_QUERY_PARAMS,
    )

    with pytest.raises(GitlabHttpError):
        gl.http_get("/not_there")
    assert responses.assert_call_count(url, 1) is True


@responses.activate
def test_get_request_invalid_data(gl):
    url = "http://localhost/api/v4/projects"
    responses.add(
        method=responses.GET,
        url=url,
        body='["name": "project1"]',
        content_type="application/json",
        status=200,
        match=helpers.MATCH_EMPTY_QUERY_PARAMS,
    )

    with pytest.raises(GitlabParsingError):
        result = gl.http_get("/projects")
        print(type(result))
        print(result.content)
    assert responses.assert_call_count(url, 1) is True


@responses.activate
def test_head_request(gl):
    url = "http://localhost/api/v4/projects"
    responses.add(
        method=responses.HEAD,
        url=url,
        headers={"X-Total": "1"},
        status=200,
        match=helpers.MATCH_EMPTY_QUERY_PARAMS,
    )

    result = gl.http_head("/projects")
    assert isinstance(result, requests.structures.CaseInsensitiveDict)
    assert result["X-Total"] == "1"


@responses.activate
def test_list_request(gl):
    url = "http://localhost/api/v4/projects"
    responses.add(
        method=responses.GET,
        url=url,
        json=[{"name": "project1"}],
        headers={"X-Total": "1"},
        status=200,
        match=helpers.MATCH_EMPTY_QUERY_PARAMS,
    )

    with warnings.catch_warnings(record=True) as caught_warnings:
        result = gl.http_list("/projects", iterator=False)
    assert len(caught_warnings) == 0
    assert isinstance(result, list)
    assert len(result) == 1

    result = gl.http_list("/projects", iterator=True)
    assert isinstance(result, GitlabList)
    assert len(list(result)) == 1

    result = gl.http_list("/projects", get_all=True)
    assert isinstance(result, list)
    assert len(result) == 1
    assert responses.assert_call_count(url, 3) is True


@responses.activate
def test_list_request_page_and_iterator(gl):
    response_dict = copy.deepcopy(large_list_response)
    response_dict["match"] = [responses.matchers.query_param_matcher({"page": "1"})]
    responses.add(**response_dict)

    with pytest.warns(
        UserWarning, match="`iterator=True` and `page=1` were both specified"
    ):
        result = gl.http_list("/projects", iterator=True, page=1)
    assert isinstance(result, list)
    assert len(result) == 20
    assert len(responses.calls) == 1


large_list_response = {
    "method": responses.GET,
    "url": "http://localhost/api/v4/projects",
    "json": [
        {"name": "project01"},
        {"name": "project02"},
        {"name": "project03"},
        {"name": "project04"},
        {"name": "project05"},
        {"name": "project06"},
        {"name": "project07"},
        {"name": "project08"},
        {"name": "project09"},
        {"name": "project10"},
        {"name": "project11"},
        {"name": "project12"},
        {"name": "project13"},
        {"name": "project14"},
        {"name": "project15"},
        {"name": "project16"},
        {"name": "project17"},
        {"name": "project18"},
        {"name": "project19"},
        {"name": "project20"},
    ],
    "headers": {"X-Total": "30", "x-per-page": "20"},
    "status": 200,
    "match": helpers.MATCH_EMPTY_QUERY_PARAMS,
}


@responses.activate
def test_list_request_pagination_warning(gl):
    responses.add(**large_list_response)

    with warnings.catch_warnings(record=True) as caught_warnings:
        result = gl.http_list("/projects", iterator=False)
    assert len(caught_warnings) == 1
    warning = caught_warnings[0]
    assert isinstance(warning.message, UserWarning)
    message = str(warning.message)
    assert "Calling a `list()` method" in message
    assert "python-gitlab.readthedocs.io" in message
    assert __file__ in message
    assert __file__ == warning.filename
    assert isinstance(result, list)
    assert len(result) == 20
    assert len(responses.calls) == 1


@responses.activate
def test_list_request_iterator_true_nowarning(gl):
    responses.add(**large_list_response)
    with warnings.catch_warnings(record=True) as caught_warnings:
        result = gl.http_list("/projects", iterator=True)
    assert len(caught_warnings) == 0
    assert isinstance(result, GitlabList)
    assert len(list(result)) == 20
    assert len(responses.calls) == 1


@responses.activate
def test_list_request_all_true_nowarning(gl):
    responses.add(**large_list_response)
    with warnings.catch_warnings(record=True) as caught_warnings:
        result = gl.http_list("/projects", get_all=True)
    assert len(caught_warnings) == 0
    assert isinstance(result, list)
    assert len(result) == 20
    assert len(responses.calls) == 1


@responses.activate
def test_list_request_all_false_nowarning(gl):
    responses.add(**large_list_response)
    with warnings.catch_warnings(record=True) as caught_warnings:
        result = gl.http_list("/projects", all=False)
    assert len(caught_warnings) == 0
    assert isinstance(result, list)
    assert len(result) == 20
    assert len(responses.calls) == 1


@responses.activate
def test_list_request_page_nowarning(gl):
    response_dict = copy.deepcopy(large_list_response)
    response_dict["match"] = [responses.matchers.query_param_matcher({"page": "1"})]
    responses.add(**response_dict)
    with warnings.catch_warnings(record=True) as caught_warnings:
        gl.http_list("/projects", page=1)
    assert len(caught_warnings) == 0
    assert len(responses.calls) == 1


@responses.activate
def test_list_request_404(gl):
    url = "http://localhost/api/v4/not_there"
    responses.add(
        method=responses.GET,
        url=url,
        json=[],
        status=404,
        match=helpers.MATCH_EMPTY_QUERY_PARAMS,
    )

    with pytest.raises(GitlabHttpError):
        gl.http_list("/not_there")
    assert responses.assert_call_count(url, 1) is True


@responses.activate
def test_list_request_invalid_data(gl):
    url = "http://localhost/api/v4/projects"
    responses.add(
        method=responses.GET,
        url=url,
        body='["name": "project1"]',
        content_type="application/json",
        status=200,
        match=helpers.MATCH_EMPTY_QUERY_PARAMS,
    )

    with pytest.raises(GitlabParsingError):
        gl.http_list("/projects")
    assert responses.assert_call_count(url, 1) is True


@responses.activate
def test_post_request(gl):
    url = "http://localhost/api/v4/projects"
    responses.add(
        method=responses.POST,
        url=url,
        json={"name": "project1"},
        status=200,
        match=helpers.MATCH_EMPTY_QUERY_PARAMS,
    )

    result = gl.http_post("/projects")
    assert isinstance(result, dict)
    assert result["name"] == "project1"
    assert responses.assert_call_count(url, 1) is True


@responses.activate
def test_post_request_404(gl):
    url = "http://localhost/api/v4/not_there"
    responses.add(
        method=responses.POST,
        url=url,
        json=[],
        status=404,
        match=helpers.MATCH_EMPTY_QUERY_PARAMS,
    )

    with pytest.raises(GitlabHttpError):
        gl.http_post("/not_there")
    assert responses.assert_call_count(url, 1) is True


@responses.activate
def test_post_request_invalid_data(gl):
    url = "http://localhost/api/v4/projects"
    responses.add(
        method=responses.POST,
        url=url,
        content_type="application/json",
        body='["name": "project1"]',
        status=200,
        match=helpers.MATCH_EMPTY_QUERY_PARAMS,
    )

    with pytest.raises(GitlabParsingError):
        gl.http_post("/projects")
    assert responses.assert_call_count(url, 1) is True


@responses.activate
def test_put_request(gl):
    url = "http://localhost/api/v4/projects"
    responses.add(
        method=responses.PUT,
        url=url,
        json={"name": "project1"},
        status=200,
        match=helpers.MATCH_EMPTY_QUERY_PARAMS,
    )

    result = gl.http_put("/projects")
    assert isinstance(result, dict)
    assert result["name"] == "project1"
    assert responses.assert_call_count(url, 1) is True


@responses.activate
def test_put_request_404(gl):
    url = "http://localhost/api/v4/not_there"
    responses.add(
        method=responses.PUT,
        url=url,
        json=[],
        status=404,
        match=helpers.MATCH_EMPTY_QUERY_PARAMS,
    )

    with pytest.raises(GitlabHttpError):
        gl.http_put("/not_there")
    assert responses.assert_call_count(url, 1) is True


@responses.activate
def test_put_request_204(gl):
    url = "http://localhost/api/v4/projects"
    responses.add(
        method=responses.PUT,
        url=url,
        status=204,
        match=helpers.MATCH_EMPTY_QUERY_PARAMS,
    )

    result = gl.http_put("/projects")
    assert isinstance(result, requests.Response)
    assert responses.assert_call_count(url, 1) is True


@responses.activate
def test_put_request_invalid_data(gl):
    url = "http://localhost/api/v4/projects"
    responses.add(
        method=responses.PUT,
        url=url,
        body='["name": "project1"]',
        content_type="application/json",
        status=200,
        match=helpers.MATCH_EMPTY_QUERY_PARAMS,
    )

    with pytest.raises(GitlabParsingError):
        gl.http_put("/projects")
    assert responses.assert_call_count(url, 1) is True


@responses.activate
def test_patch_request(gl):
    url = "http://localhost/api/v4/projects"
    responses.add(
        method=responses.PATCH,
        url=url,
        json={"name": "project1"},
        status=200,
        match=helpers.MATCH_EMPTY_QUERY_PARAMS,
    )

    result = gl.http_patch("/projects")
    assert isinstance(result, dict)
    assert result["name"] == "project1"
    assert responses.assert_call_count(url, 1) is True


@responses.activate
def test_patch_request_204(gl):
    url = "http://localhost/api/v4/projects"
    responses.add(
        method=responses.PATCH,
        url=url,
        status=204,
        match=helpers.MATCH_EMPTY_QUERY_PARAMS,
    )

    result = gl.http_patch("/projects")
    assert isinstance(result, requests.Response)
    assert responses.assert_call_count(url, 1) is True


@responses.activate
def test_patch_request_404(gl):
    url = "http://localhost/api/v4/not_there"
    responses.add(
        method=responses.PATCH,
        url=url,
        json=[],
        status=404,
        match=helpers.MATCH_EMPTY_QUERY_PARAMS,
    )

    with pytest.raises(GitlabHttpError):
        gl.http_patch("/not_there")
    assert responses.assert_call_count(url, 1) is True


@responses.activate
def test_patch_request_invalid_data(gl):
    url = "http://localhost/api/v4/projects"
    responses.add(
        method=responses.PATCH,
        url=url,
        body='["name": "project1"]',
        content_type="application/json",
        status=200,
        match=helpers.MATCH_EMPTY_QUERY_PARAMS,
    )

    with pytest.raises(GitlabParsingError):
        gl.http_patch("/projects")
    assert responses.assert_call_count(url, 1) is True


@responses.activate
def test_delete_request(gl):
    url = "http://localhost/api/v4/projects"
    responses.add(
        method=responses.DELETE,
        url=url,
        json=True,
        status=200,
        match=helpers.MATCH_EMPTY_QUERY_PARAMS,
    )

    result = gl.http_delete("/projects")
    assert isinstance(result, requests.Response)
    assert result.json() is True
    assert responses.assert_call_count(url, 1) is True


@responses.activate
def test_delete_request_404(gl):
    url = "http://localhost/api/v4/not_there"
    responses.add(
        method=responses.DELETE,
        url=url,
        json=[],
        status=404,
        match=helpers.MATCH_EMPTY_QUERY_PARAMS,
    )

    with pytest.raises(GitlabHttpError):
        gl.http_delete("/not_there")
    assert responses.assert_call_count(url, 1) is True



"""
GitLab API: https://docs.gitlab.com/ce/api/appearance.html
"""

import pytest
import responses

title = "GitLab Test Instance"
description = "gitlab-test.example.com"
new_title = "new-title"
new_description = "new-description"


@pytest.fixture
def resp_application_appearance():
    content = {
        "title": title,
        "description": description,
        "logo": "/uploads/-/system/appearance/logo/1/logo.png",
        "header_logo": "/uploads/-/system/appearance/header_logo/1/header.png",
        "favicon": "/uploads/-/system/appearance/favicon/1/favicon.png",
        "new_project_guidelines": "Please read the FAQs for help.",
        "header_message": "",
        "footer_message": "",
        "message_background_color": "#e75e40",
        "message_font_color": "#ffffff",
        "email_header_and_footer_enabled": False,
    }

    with responses.RequestsMock(assert_all_requests_are_fired=False) as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/application/appearance",
            json=content,
            content_type="application/json",
            status=200,
        )

        updated_content = dict(content)
        updated_content["title"] = new_title
        updated_content["description"] = new_description

        rsps.add(
            method=responses.PUT,
            url="http://localhost/api/v4/application/appearance",
            json=updated_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


def test_get_update_appearance(gl, resp_application_appearance):
    appearance = gl.appearance.get()
    assert appearance.title == title
    assert appearance.description == description
    appearance.title = new_title
    appearance.description = new_description
    appearance.save()
    assert appearance.title == new_title
    assert appearance.description == new_description


def test_update_appearance(gl, resp_application_appearance):
    gl.appearance.update(title=new_title, description=new_description)



"""
GitLab API: https://docs.gitlab.com/ee/api/statistics.html
"""

import pytest
import responses

content = {
    "forks": "10",
    "issues": "76",
    "merge_requests": "27",
    "notes": "954",
    "snippets": "50",
    "ssh_keys": "10",
    "milestones": "40",
    "users": "50",
    "groups": "10",
    "projects": "20",
    "active_users": "50",
}


@pytest.fixture
def resp_application_statistics():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/application/statistics",
            json=content,
            content_type="application/json",
            status=200,
        )

        yield rsps


def test_get_statistics(gl, resp_application_statistics):
    statistics = gl.statistics.get()
    assert statistics.forks == content["forks"]
    assert statistics.merge_requests == content["merge_requests"]
    assert statistics.notes == content["notes"]
    assert statistics.snippets == content["snippets"]
    assert statistics.ssh_keys == content["ssh_keys"]
    assert statistics.milestones == content["milestones"]
    assert statistics.users == content["users"]
    assert statistics.groups == content["groups"]
    assert statistics.projects == content["projects"]
    assert statistics.active_users == content["active_users"]



"""
GitLab API: https://docs.gitlab.com/ee/api/pipelines.html
"""

import pytest
import responses

from gitlab.v4.objects import (
    ProjectPipeline,
    ProjectPipelineTestReport,
    ProjectPipelineTestReportSummary,
)

pipeline_content = {
    "id": 46,
    "project_id": 1,
    "status": "pending",
    "ref": "main",
    "sha": "a91957a858320c0e17f3a0eca7cfacbff50ea29a",
    "before_sha": "a91957a858320c0e17f3a0eca7cfacbff50ea29a",
    "tag": False,
    "yaml_errors": None,
    "user": {
        "name": "Administrator",
        "username": "root",
        "id": 1,
        "state": "active",
        "avatar_url": "http://www.gravatar.com/avatar/e64c7d89f26bd1972efa854d13d7dd61?s=80&d=identicon",
        "web_url": "http://localhost:3000/root",
    },
    "created_at": "2016-08-11T11:28:34.085Z",
    "updated_at": "2016-08-11T11:32:35.169Z",
    "started_at": None,
    "finished_at": "2016-08-11T11:32:35.145Z",
    "committed_at": None,
    "duration": None,
    "queued_duration": 0.010,
    "coverage": None,
    "web_url": "https://example.com/foo/bar/pipelines/46",
}

pipeline_latest = {
    "id": 47,
    "project_id": 1,
    "status": "pending",
    "ref": "main",
    "sha": "a91957a858320c0e17f3a0eca7cfacbff50ea29a",
    "before_sha": "a91957a858320c0e17f3a0eca7cfacbff50ea29a",
    "tag": False,
    "yaml_errors": None,
    "user": {
        "name": "Administrator",
        "username": "root",
        "id": 1,
        "state": "active",
        "avatar_url": "http://www.gravatar.com/avatar/e64c7d89f26bd1972efa854d13d7dd61?s=80&d=identicon",
        "web_url": "http://localhost:3000/root",
    },
    "created_at": "2016-08-11T11:28:34.085Z",
    "updated_at": "2016-08-11T11:32:35.169Z",
    "started_at": None,
    "finished_at": "2016-08-11T11:32:35.145Z",
    "committed_at": None,
    "duration": None,
    "queued_duration": 0.010,
    "coverage": None,
    "web_url": "https://example.com/foo/bar/pipelines/46",
}

pipeline_latest_other_ref = {
    "id": 48,
    "project_id": 1,
    "status": "pending",
    "ref": "feature-ref",
    "sha": "a91957a858320c0e17f3a0eca7cfacbff50ea29a",
    "before_sha": "a91957a858320c0e17f3a0eca7cfacbff50ea29a",
    "tag": False,
    "yaml_errors": None,
    "user": {
        "name": "Administrator",
        "username": "root",
        "id": 1,
        "state": "active",
        "avatar_url": "http://www.gravatar.com/avatar/e64c7d89f26bd1972efa854d13d7dd61?s=80&d=identicon",
        "web_url": "http://localhost:3000/root",
    },
    "created_at": "2016-08-11T11:28:34.085Z",
    "updated_at": "2016-08-11T11:32:35.169Z",
    "started_at": None,
    "finished_at": "2016-08-11T11:32:35.145Z",
    "committed_at": None,
    "duration": None,
    "queued_duration": 0.010,
    "coverage": None,
    "web_url": "https://example.com/foo/bar/pipelines/46",
}


test_report_content = {
    "total_time": 5,
    "total_count": 1,
    "success_count": 1,
    "failed_count": 0,
    "skipped_count": 0,
    "error_count": 0,
    "test_suites": [
        {
            "name": "Secure",
            "total_time": 5,
            "total_count": 1,
            "success_count": 1,
            "failed_count": 0,
            "skipped_count": 0,
            "error_count": 0,
            "test_cases": [
                {
                    "status": "success",
                    "name": "Security Reports can create an auto-remediation MR",
                    "classname": "vulnerability_management_spec",
                    "execution_time": 5,
                    "system_output": None,
                    "stack_trace": None,
                }
            ],
        }
    ],
}


test_report_summary_content = {
    "total": {
        "time": 1904,
        "count": 3363,
        "success": 3351,
        "failed": 0,
        "skipped": 12,
        "error": 0,
        "suite_error": None,
    },
    "test_suites": [
        {
            "name": "test",
            "total_time": 1904,
            "total_count": 3363,
            "success_count": 3351,
            "failed_count": 0,
            "skipped_count": 12,
            "error_count": 0,
            "build_ids": [66004],
            "suite_error": None,
        }
    ],
}


@pytest.fixture
def resp_get_pipeline():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/projects/1/pipelines/1",
            json=pipeline_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_cancel_pipeline():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.POST,
            url="http://localhost/api/v4/projects/1/pipelines/1/cancel",
            json=pipeline_content,
            content_type="application/json",
            status=201,
        )
        yield rsps


@pytest.fixture
def resp_retry_pipeline():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.POST,
            url="http://localhost/api/v4/projects/1/pipelines/1/retry",
            json=pipeline_content,
            content_type="application/json",
            status=201,
        )
        yield rsps


@pytest.fixture
def resp_get_pipeline_test_report():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/projects/1/pipelines/1/test_report",
            json=test_report_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_get_pipeline_test_report_summary():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/projects/1/pipelines/1/test_report_summary",
            json=test_report_summary_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_get_latest():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/projects/1/pipelines/latest",
            json=pipeline_latest,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_get_latest_other_ref():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/projects/1/pipelines/latest",
            json=pipeline_latest_other_ref,
            content_type="application/json",
            status=200,
        )
        yield rsps


def test_get_project_pipeline(project, resp_get_pipeline):
    pipeline = project.pipelines.get(1)
    assert isinstance(pipeline, ProjectPipeline)
    assert pipeline.ref == "main"
    assert pipeline.id == 46


def test_cancel_project_pipeline(project, resp_cancel_pipeline):
    pipeline = project.pipelines.get(1, lazy=True)

    output = pipeline.cancel()
    assert output["ref"] == "main"


def test_retry_project_pipeline(project, resp_retry_pipeline):
    pipeline = project.pipelines.get(1, lazy=True)

    output = pipeline.retry()
    assert output["ref"] == "main"


def test_get_project_pipeline_test_report(project, resp_get_pipeline_test_report):
    pipeline = project.pipelines.get(1, lazy=True)
    test_report = pipeline.test_report.get()
    assert isinstance(test_report, ProjectPipelineTestReport)
    assert test_report.total_time == 5
    assert test_report.test_suites[0]["name"] == "Secure"


def test_get_project_pipeline_test_report_summary(
    project, resp_get_pipeline_test_report_summary
):
    pipeline = project.pipelines.get(1, lazy=True)
    test_report_summary = pipeline.test_report_summary.get()
    assert isinstance(test_report_summary, ProjectPipelineTestReportSummary)
    assert test_report_summary.total["count"] == 3363
    assert test_report_summary.test_suites[0]["name"] == "test"


def test_latest_pipeline(project, resp_get_latest):
    pipeline = project.pipelines.latest()
    assert isinstance(pipeline, ProjectPipeline)
    assert pipeline.ref == "main"
    assert pipeline.id == 47


def test_latest_pipeline_other_ref(project, resp_get_latest_other_ref):
    pipeline = project.pipelines.latest(ref="feature-ref")
    assert isinstance(pipeline, ProjectPipeline)
    assert pipeline.ref == "feature-ref"
    assert pipeline.id == 48



import json
import logging
import warnings

import pytest
import requests
import responses

from gitlab import types, utils


@pytest.mark.parametrize(
    "content_type,expected_type",
    [
        ("application/json", "application/json"),
        ("application/json; charset=utf-8", "application/json"),
        ("", "text/plain"),
        (None, "text/plain"),
    ],
)
def test_get_content_type(content_type, expected_type):
    parsed_type = utils.get_content_type(content_type)
    assert parsed_type == expected_type


@responses.activate
def test_response_content(capsys):
    responses.add(
        method="GET",
        url="https://example.com",
        status=200,
        body="test",
        content_type="application/octet-stream",
    )

    resp = requests.get("https://example.com", stream=True)
    utils.response_content(
        resp, streamed=True, action=None, chunk_size=1024, iterator=False
    )

    captured = capsys.readouterr()
    assert "test" in captured.out


class TestEncodedId:
    def test_init_str(self):
        obj = utils.EncodedId("Hello")
        assert "Hello" == obj
        assert "Hello" == str(obj)
        assert "Hello" == f"{obj}"
        assert isinstance(obj, utils.EncodedId)

        obj = utils.EncodedId("this/is a/path")
        assert "this%2Fis%20a%2Fpath" == str(obj)
        assert "this%2Fis%20a%2Fpath" == f"{obj}"
        assert isinstance(obj, utils.EncodedId)

    def test_init_int(self):
        obj = utils.EncodedId(23)
        assert "23" == obj
        assert "23" == f"{obj}"
        assert isinstance(obj, utils.EncodedId)

    def test_init_invalid_type_raises(self):
        with pytest.raises(TypeError):
            utils.EncodedId(None)

    def test_init_encodeid_str(self):
        value = "Goodbye"
        obj_init = utils.EncodedId(value)
        obj = utils.EncodedId(obj_init)
        assert value == str(obj)
        assert value == f"{obj}"

        value = "we got/a/path"
        expected = "we%20got%2Fa%2Fpath"
        obj_init = utils.EncodedId(value)
        assert expected == str(obj_init)
        assert expected == f"{obj_init}"
        # Show that no matter how many times we recursively call it we still only
        # URL-encode it once.
        obj = utils.EncodedId(
            utils.EncodedId(utils.EncodedId(utils.EncodedId(utils.EncodedId(obj_init))))
        )
        assert expected == str(obj)
        assert expected == f"{obj}"

        # Show assignments still only encode once
        obj2 = obj
        assert expected == str(obj2)
        assert expected == f"{obj2}"

    def test_init_encodeid_int(self):
        value = 23
        expected = f"{value}"
        obj_init = utils.EncodedId(value)
        obj = utils.EncodedId(obj_init)
        assert expected == str(obj)
        assert expected == f"{obj}"

    def test_json_serializable(self):
        obj = utils.EncodedId("someone")
        assert '"someone"' == json.dumps(obj)

        obj = utils.EncodedId("we got/a/path")
        assert '"we%20got%2Fa%2Fpath"' == json.dumps(obj)


class TestWarningsWrapper:
    def test_warn(self):
        warn_message = "short and stout"
        warn_source = "teapot"

        with warnings.catch_warnings(record=True) as caught_warnings:
            utils.warn(message=warn_message, category=UserWarning, source=warn_source)
        assert len(caught_warnings) == 1
        warning = caught_warnings[0]
        # File name is this file as it is the first file outside of the `gitlab/` path.
        assert __file__ == warning.filename
        assert warning.category == UserWarning
        assert isinstance(warning.message, UserWarning)
        assert warn_message in str(warning.message)
        assert __file__ in str(warning.message)
        assert warn_source == warning.source

    def test_warn_no_show_caller(self):
        warn_message = "short and stout"
        warn_source = "teapot"

        with warnings.catch_warnings(record=True) as caught_warnings:
            utils.warn(
                message=warn_message,
                category=UserWarning,
                source=warn_source,
                show_caller=False,
            )
        assert len(caught_warnings) == 1
        warning = caught_warnings[0]
        # File name is this file as it is the first file outside of the `gitlab/` path.
        assert __file__ == warning.filename
        assert warning.category == UserWarning
        assert isinstance(warning.message, UserWarning)
        assert warn_message in str(warning.message)
        assert __file__ not in str(warning.message)
        assert warn_source == warning.source


@pytest.mark.parametrize(
    "source,expected",
    [
        ({"a": "", "b": "spam", "c": None}, {"a": "", "b": "spam", "c": None}),
        ({"a": "", "b": {"c": "spam"}}, {"a": "", "b[c]": "spam"}),
    ],
)
def test_copy_dict(source, expected):
    dest = {}

    utils.copy_dict(src=source, dest=dest)
    assert dest == expected


@pytest.mark.parametrize(
    "dictionary,expected",
    [
        ({"a": None, "b": "spam"}, {"b": "spam"}),
        ({"a": "", "b": "spam"}, {"a": "", "b": "spam"}),
        ({"a": None, "b": None}, {}),
    ],
)
def test_remove_none_from_dict(dictionary, expected):
    result = utils.remove_none_from_dict(dictionary)
    assert result == expected


def test_transform_types_copies_data_with_empty_files():
    data = {"attr": "spam"}
    new_data, files = utils._transform_types(data, {}, transform_data=True)

    assert new_data is not data
    assert new_data == data
    assert files == {}


def test_transform_types_with_transform_files_populates_files():
    custom_types = {"attr": types.FileAttribute}
    data = {"attr": "spam"}
    new_data, files = utils._transform_types(data, custom_types, transform_data=True)

    assert new_data == {}
    assert files["attr"] == ("attr", "spam")


def test_transform_types_without_transform_files_populates_data_with_empty_files():
    custom_types = {"attr": types.FileAttribute}
    data = {"attr": "spam"}
    new_data, files = utils._transform_types(
        data, custom_types, transform_files=False, transform_data=True
    )

    assert new_data == {"attr": "spam"}
    assert files == {}


def test_transform_types_params_array():
    data = {"attr": [1, 2, 3]}
    custom_types = {"attr": types.ArrayAttribute}
    new_data, files = utils._transform_types(data, custom_types, transform_data=True)

    assert new_data is not data
    assert new_data == {"attr[]": [1, 2, 3]}
    assert files == {}


def test_transform_types_not_params_array():
    data = {"attr": [1, 2, 3]}
    custom_types = {"attr": types.ArrayAttribute}
    new_data, files = utils._transform_types(data, custom_types, transform_data=False)

    assert new_data is not data
    assert new_data == data
    assert files == {}


def test_masking_formatter_masks_token(capsys: pytest.CaptureFixture):
    token = "glpat-private-token"

    logger = logging.getLogger()
    handler = logging.StreamHandler()
    handler.setFormatter(utils.MaskingFormatter(masked=token))
    logger.handlers.clear()
    logger.addHandler(handler)

    logger.info(token)
    captured = capsys.readouterr()

    assert "[MASKED]" in captured.err
    assert token not in captured.err



import pytest
import responses

from gitlab.const import DEFAULT_URL


@pytest.fixture
def gitlab_cli(script_runner, gitlab_config):
    """Wrapper fixture to help make test cases less verbose."""

    def _gitlab_cli(subcommands):
        """
        Return a script_runner.run method that takes a default gitlab
        command, and subcommands passed as arguments inside test cases.
        """
        command = ["gitlab", "--config-file", gitlab_config]

        for subcommand in subcommands:
            # ensure we get strings (e.g from IDs)
            command.append(str(subcommand))

        return script_runner.run(command)

    return _gitlab_cli


@pytest.fixture
def resp_get_project():
    return {
        "method": responses.GET,
        "url": f"{DEFAULT_URL}/api/v4/projects/1",
        "json": {"name": "name", "path": "test-path", "id": 1},
        "content_type": "application/json",
        "status": 200,
    }


@pytest.fixture
def resp_current_user():
    return {
        "method": responses.GET,
        "url": f"{DEFAULT_URL}/api/v4/user",
        "json": {"username": "name", "id": 1},
        "content_type": "application/json",
        "status": 200,
    }


@pytest.fixture
def resp_delete_registry_tags_in_bulk():
    return {
        "method": responses.DELETE,
        "url": f"{DEFAULT_URL}/api/v4/projects/1/registry/repositories/1/tags",
        "status": 202,
    }



import pytest

import gitlab


def test_snippets(gl):
    snippets = gl.snippets.list(get_all=True)
    assert not snippets

    snippet = gl.snippets.create(
        {
            "title": "snippet1",
            "files": [{"file_path": "snippet1.py", "content": "import gitlab"}],
        }
    )
    snippet = gl.snippets.get(snippet.id)
    snippet.title = "updated_title"
    snippet.save()

    snippet = gl.snippets.get(snippet.id)
    assert snippet.title == "updated_title"

    content = snippet.content()
    assert content.decode() == "import gitlab"

    all_snippets = gl.snippets.list_all(get_all=True)
    public_snippets = gl.snippets.public(get_all=True)
    list_public_snippets = gl.snippets.list_public(get_all=True)
    assert isinstance(all_snippets, list)
    assert isinstance(list_public_snippets, list)
    assert public_snippets == list_public_snippets

    snippet.delete()


def test_project_snippets(project):
    project.snippets_enabled = True
    project.save()

    snippet = project.snippets.create(
        {
            "title": "snip1",
            "files": [{"file_path": "foo.py", "content": "initial content"}],
            "visibility": gitlab.const.VISIBILITY_PRIVATE,
        }
    )

    assert snippet.title == "snip1"


@pytest.mark.xfail(reason="Returning 404 UserAgentDetail not found in GL 16")
def test_project_snippet_user_agent_detail(project):
    snippet = project.snippets.list()[0]

    user_agent_detail = snippet.user_agent_detail()

    assert user_agent_detail["user_agent"]


def test_project_snippet_discussion(project):
    snippet = project.snippets.list()[0]

    discussion = snippet.discussions.create({"body": "Discussion body"})
    assert discussion in snippet.discussions.list()

    note = discussion.notes.create({"body": "first note"})
    note_from_get = discussion.notes.get(note.id)
    note_from_get.body = "updated body"
    note_from_get.save()

    discussion = snippet.discussions.get(discussion.id)
    assert discussion.attributes["notes"][-1]["body"] == "updated body"

    note_from_get.delete()


def test_project_snippet_file(project):
    snippet = project.snippets.list()[0]
    snippet.file_name = "bar.py"
    snippet.save()

    snippet = project.snippets.get(snippet.id)
    assert snippet.content().decode() == "initial content"
    assert snippet.file_name == "bar.py"
    assert snippet in project.snippets.list()

    snippet.delete()



"""
GitLab API: https://docs.gitlab.com/ee/api/members.html
"""

import pytest
import responses

from gitlab.const import AccessLevel
from gitlab.v4.objects import GroupBillableMember

billable_members_content = [
    {
        "id": 1,
        "username": "raymond_smith",
        "name": "Raymond Smith",
        "state": "active",
        "avatar_url": "https://www.gravatar.com/avatar/c2525a7f58ae3776070e44c106c48e15?s=80&d=identicon",
        "web_url": "http://192.168.1.8:3000/root",
        "last_activity_on": "2021-01-27",
        "membership_type": "group_member",
        "removable": True,
    }
]


@pytest.fixture
def resp_create_group_member():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.POST,
            url="http://localhost/api/v4/groups/1/members",
            json={"id": 1, "username": "jane_doe", "access_level": 30},
            content_type="application/json",
            status=201,
        )
        yield rsps


@pytest.fixture
def resp_list_billable_group_members():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/groups/1/billable_members",
            json=billable_members_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_delete_billable_group_member():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.DELETE,
            url="http://localhost/api/v4/groups/1/billable_members/1",
            status=204,
        )
        yield rsps


def test_create_group_member(group, resp_create_group_member):
    member = group.members.create({"user_id": 1, "access_level": AccessLevel.DEVELOPER})
    assert member.access_level == 30


def test_list_group_billable_members(group, resp_list_billable_group_members):
    billable_members = group.billable_members.list()
    assert isinstance(billable_members, list)
    assert isinstance(billable_members[0], GroupBillableMember)
    assert billable_members[0].removable is True


def test_delete_group_billable_member(group, resp_delete_billable_group_member):
    group.billable_members.delete(1)



"""
GitLab API:
https://docs.gitlab.com/ee/api/instance_level_ci_variables.html
https://docs.gitlab.com/ee/api/project_level_variables.html
https://docs.gitlab.com/ee/api/group_level_variables.html
"""

import re

import pytest
import responses

from gitlab.v4.objects import GroupVariable, ProjectVariable, Variable

key = "TEST_VARIABLE_1"
value = "TEST_1"
new_value = "TEST_2"

variable_content = {
    "key": key,
    "variable_type": "env_var",
    "value": value,
    "protected": False,
    "masked": True,
}
variables_url = re.compile(
    r"http://localhost/api/v4/(((groups|projects)/1)|(admin/ci))/variables"
)
variables_key_url = re.compile(
    rf"http://localhost/api/v4/(((groups|projects)/1)|(admin/ci))/variables/{key}"
)


@pytest.fixture
def resp_list_variables():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url=variables_url,
            json=[variable_content],
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_get_variable():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url=variables_key_url,
            json=variable_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_create_variable():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.POST,
            url=variables_url,
            json=variable_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_update_variable():
    updated_content = dict(variable_content)
    updated_content["value"] = new_value

    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.PUT,
            url=variables_key_url,
            json=updated_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_delete_variable():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.DELETE,
            url=variables_key_url,
            status=204,
        )
        yield rsps


def test_list_instance_variables(gl, resp_list_variables):
    variables = gl.variables.list()
    assert isinstance(variables, list)
    assert isinstance(variables[0], Variable)
    assert variables[0].value == value


def test_get_instance_variable(gl, resp_get_variable):
    variable = gl.variables.get(key)
    assert isinstance(variable, Variable)
    assert variable.value == value


def test_create_instance_variable(gl, resp_create_variable):
    variable = gl.variables.create({"key": key, "value": value})
    assert isinstance(variable, Variable)
    assert variable.value == value


def test_update_instance_variable(gl, resp_update_variable):
    variable = gl.variables.get(key, lazy=True)
    variable.value = new_value
    variable.save()
    assert variable.value == new_value


def test_delete_instance_variable(gl, resp_delete_variable):
    variable = gl.variables.get(key, lazy=True)
    variable.delete()


def test_list_project_variables(project, resp_list_variables):
    variables = project.variables.list()
    assert isinstance(variables, list)
    assert isinstance(variables[0], ProjectVariable)
    assert variables[0].value == value


def test_get_project_variable(project, resp_get_variable):
    variable = project.variables.get(key)
    assert isinstance(variable, ProjectVariable)
    assert variable.value == value


def test_create_project_variable(project, resp_create_variable):
    variable = project.variables.create({"key": key, "value": value})
    assert isinstance(variable, ProjectVariable)
    assert variable.value == value


def test_update_project_variable(project, resp_update_variable):
    variable = project.variables.get(key, lazy=True)
    variable.value = new_value
    variable.save()
    assert variable.value == new_value


def test_delete_project_variable(project, resp_delete_variable):
    variable = project.variables.get(key, lazy=True)
    variable.delete()


def test_list_group_variables(group, resp_list_variables):
    variables = group.variables.list()
    assert isinstance(variables, list)
    assert isinstance(variables[0], GroupVariable)
    assert variables[0].value == value


def test_get_group_variable(group, resp_get_variable):
    variable = group.variables.get(key)
    assert isinstance(variable, GroupVariable)
    assert variable.value == value


def test_create_group_variable(group, resp_create_variable):
    variable = group.variables.create({"key": key, "value": value})
    assert isinstance(variable, GroupVariable)
    assert variable.value == value


def test_update_group_variable(group, resp_update_variable):
    variable = group.variables.get(key, lazy=True)
    variable.value = new_value
    variable.save()
    assert variable.value == new_value


def test_delete_group_variable(group, resp_delete_variable):
    variable = group.variables.get(key, lazy=True)
    variable.delete()



from gitlab import base
from tests.unit import helpers


def test_computed_path_simple(gl):
    class MGR(base.RESTManager):
        _path = "/tests"
        _obj_cls = object

    mgr = MGR(gl)
    assert mgr._computed_path == "/tests"


def test_computed_path_with_parent(gl, fake_manager):
    class MGR(base.RESTManager):
        _path = "/tests/{test_id}/cases"
        _obj_cls = object
        _from_parent_attrs = {"test_id": "id"}

    mgr = MGR(gl, parent=helpers.FakeParent(manager=fake_manager, attrs={}))
    assert mgr._computed_path == "/tests/42/cases"


def test_path_property(gl):
    class MGR(base.RESTManager):
        _path = "/tests"
        _obj_cls = object

    mgr = MGR(gl)
    assert mgr.path == "/tests"



"""
GitLab API: https://docs.gitlab.com/ee/api/secure_files.html
"""

import pytest
import responses

from gitlab.v4.objects import ProjectSecureFile

secure_file_content = {
    "id": 1,
    "name": "myfile.jks",
    "checksum": "16630b189ab34b2e3504f4758e1054d2e478deda510b2b08cc0ef38d12e80aac",
    "checksum_algorithm": "sha256",
    "created_at": "2022-02-22T22:22:22.222Z",
    "expires_at": None,
    "metadata": None,
}


@pytest.fixture
def resp_list_secure_files():
    with responses.RequestsMock(assert_all_requests_are_fired=False) as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/projects/1/secure_files",
            json=[secure_file_content],
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_create_secure_file():
    with responses.RequestsMock(assert_all_requests_are_fired=False) as rsps:
        rsps.add(
            method=responses.POST,
            url="http://localhost/api/v4/projects/1/secure_files",
            json=secure_file_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_download_secure_file(binary_content):
    with responses.RequestsMock(assert_all_requests_are_fired=False) as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/projects/1/secure_files/1",
            json=secure_file_content,
            content_type="application/json",
            status=200,
        )
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/projects/1/secure_files/1/download",
            body=binary_content,
            content_type="application/octet-stream",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_remove_secure_file():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.DELETE,
            url="http://localhost/api/v4/projects/1/secure_files/1",
            status=204,
        )
        yield rsps


def test_list_secure_files(project, resp_list_secure_files):
    secure_files = project.secure_files.list()
    assert len(secure_files) == 1
    assert secure_files[0].id == 1
    assert secure_files[0].name == "myfile.jks"


def test_create_secure_file(project, resp_create_secure_file):
    secure_files = project.secure_files.create({"name": "test", "file": "myfile.jks"})
    assert secure_files.id == 1
    assert secure_files.name == "myfile.jks"


def test_download_secure_file(project, binary_content, resp_download_secure_file):
    secure_file = project.secure_files.get(1)
    secure_content = secure_file.download()
    assert isinstance(secure_file, ProjectSecureFile)
    assert secure_content == binary_content


def test_remove_secure_file(project, resp_remove_secure_file):
    project.secure_files.delete(1)



"""
GitLab API: https://docs.gitlab.com/ce/api/deployments.html
"""

import pytest
import responses


@pytest.fixture
def resp_deployment_get():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/projects/1/deployments/42",
            json=response_get_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def deployment(project):
    return project.deployments.get(42, lazy=True)


@pytest.fixture
def resp_deployment_create():
    content = {"id": 42, "status": "success", "ref": "main"}

    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.POST,
            url="http://localhost/api/v4/projects/1/deployments",
            json=content,
            content_type="application/json",
            status=200,
        )

        updated_content = dict(content)
        updated_content["status"] = "failed"

        rsps.add(
            method=responses.PUT,
            url="http://localhost/api/v4/projects/1/deployments/42",
            json=updated_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_deployment_approval():
    content = {
        "user": {
            "id": 100,
            "username": "security-user-1",
            "name": "security user-1",
            "state": "active",
            "avatar_url": "https://www.gravatar.com/avatar/e130fcd3a1681f41a3de69d10841afa9?s=80&d=identicon",
            "web_url": "http://localhost:3000/security-user-1",
        },
        "status": "approved",
        "created_at": "2022-02-24T20:22:30.097Z",
        "comment": "Looks good to me",
    }

    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.POST,
            url="http://localhost/api/v4/projects/1/deployments/42/approval",
            json=content,
            content_type="application/json",
            status=200,
        )
        yield rsps


def test_deployment_get(project, resp_deployment_get):
    deployment = project.deployments.get(42)
    assert deployment.id == 42
    assert deployment.iid == 2
    assert deployment.status == "success"
    assert deployment.ref == "main"


def test_deployment_create(project, resp_deployment_create):
    deployment = project.deployments.create(
        {
            "environment": "Test",
            "sha": "1agf4gs",
            "ref": "main",
            "tag": False,
            "status": "created",
        }
    )
    assert deployment.id == 42
    assert deployment.status == "success"
    assert deployment.ref == "main"

    deployment.status = "failed"
    deployment.save()
    assert deployment.status == "failed"


def test_deployment_approval(deployment, resp_deployment_approval) -> None:
    result = deployment.approval(status="approved")
    assert result["status"] == "approved"
    assert result["comment"] == "Looks good to me"


response_get_content = {
    "id": 42,
    "iid": 2,
    "ref": "main",
    "sha": "a91957a858320c0e17f3a0eca7cfacbff50ea29a",
    "created_at": "2016-08-11T11:32:35.444Z",
    "updated_at": "2016-08-11T11:34:01.123Z",
    "status": "success",
    "user": {
        "name": "Administrator",
        "username": "root",
        "id": 1,
        "state": "active",
        "avatar_url": "http://www.gravatar.com/avatar/e64c7d89f26bd1972efa854d13d7dd61?s=80&d=identicon",
        "web_url": "http://localhost:3000/root",
    },
    "environment": {
        "id": 9,
        "name": "production",
        "external_url": "https://about.gitlab.com",
    },
    "deployable": {
        "id": 664,
        "status": "success",
        "stage": "deploy",
        "name": "deploy",
        "ref": "main",
        "tag": False,
        "coverage": None,
        "created_at": "2016-08-11T11:32:24.456Z",
        "started_at": None,
        "finished_at": "2016-08-11T11:32:35.145Z",
        "user": {
            "id": 1,
            "name": "Administrator",
            "username": "root",
            "state": "active",
            "avatar_url": "http://www.gravatar.com/avatar/e64c7d89f26bd1972efa854d13d7dd61?s=80&d=identicon",
            "web_url": "http://gitlab.dev/root",
            "created_at": "2015-12-21T13:14:24.077Z",
            "bio": None,
            "location": None,
            "skype": "",
            "linkedin": "",
            "twitter": "",
            "website_url": "",
            "organization": "",
        },
        "commit": {
            "id": "a91957a858320c0e17f3a0eca7cfacbff50ea29a",
            "short_id": "a91957a8",
            "title": "Merge branch 'rename-readme' into 'main'\r",
            "author_name": "Administrator",
            "author_email": "admin@example.com",
            "created_at": "2016-08-11T13:28:26.000+02:00",
            "message": "Merge branch 'rename-readme' into 'main'\r\n\r\nRename README\r\n\r\n\r\n\r\nSee merge request !2",
        },
        "pipeline": {
            "created_at": "2016-08-11T07:43:52.143Z",
            "id": 42,
            "ref": "main",
            "sha": "a91957a858320c0e17f3a0eca7cfacbff50ea29a",
            "status": "success",
            "updated_at": "2016-08-11T07:43:52.143Z",
            "web_url": "http://gitlab.dev/root/project/pipelines/5",
        },
        "runner": None,
    },
}



"""
GitLab API:
https://docs.gitlab.com/ee/api/audit_events.html#project-audit-events
"""

import re

import pytest
import responses

from gitlab.v4.objects.audit_events import (
    AuditEvent,
    GroupAuditEvent,
    ProjectAuditEvent,
)

id = 5

audit_events_content = {
    "id": 5,
    "author_id": 1,
    "entity_id": 7,
    "entity_type": "Project",
    "details": {
        "change": "prevent merge request approval from reviewers",
        "from": "",
        "to": "true",
        "author_name": "Administrator",
        "target_id": 7,
        "target_type": "Project",
        "target_details": "twitter/typeahead-js",
        "ip_address": "127.0.0.1",
        "entity_path": "twitter/typeahead-js",
    },
    "created_at": "2020-05-26T22:55:04.230Z",
}

audit_events_url = re.compile(
    r"http://localhost/api/v4/((groups|projects)/1/)?audit_events"
)

audit_events_url_id = re.compile(
    rf"http://localhost/api/v4/((groups|projects)/1/)?audit_events/{id}"
)


@pytest.fixture
def resp_list_audit_events():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url=audit_events_url,
            json=[audit_events_content],
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_get_audit_event():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url=audit_events_url_id,
            json=audit_events_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


def test_list_instance_audit_events(gl, resp_list_audit_events):
    audit_events = gl.audit_events.list()
    assert isinstance(audit_events, list)
    assert isinstance(audit_events[0], AuditEvent)
    assert audit_events[0].id == id


def test_get_instance_audit_events(gl, resp_get_audit_event):
    audit_event = gl.audit_events.get(id)
    assert isinstance(audit_event, AuditEvent)
    assert audit_event.id == id


def test_list_group_audit_events(group, resp_list_audit_events):
    audit_events = group.audit_events.list()
    assert isinstance(audit_events, list)
    assert isinstance(audit_events[0], GroupAuditEvent)
    assert audit_events[0].id == id


def test_get_group_audit_events(group, resp_get_audit_event):
    audit_event = group.audit_events.get(id)
    assert isinstance(audit_event, GroupAuditEvent)
    assert audit_event.id == id


def test_list_project_audit_events(project, resp_list_audit_events):
    audit_events = project.audit_events.list()
    assert isinstance(audit_events, list)
    assert isinstance(audit_events[0], ProjectAuditEvent)
    assert audit_events[0].id == id


def test_get_project_audit_events(project, resp_get_audit_event):
    audit_event = project.audit_events.get(id)
    assert isinstance(audit_event, ProjectAuditEvent)
    assert audit_event.id == id



from typing import Any, cast, Optional, TYPE_CHECKING, Union

from gitlab.base import RESTManager, RESTObject
from gitlab.mixins import GetMixin

__all__ = [
    "Key",
    "KeyManager",
]


class Key(RESTObject):
    pass


class KeyManager(GetMixin, RESTManager):
    _path = "/keys"
    _obj_cls = Key

    def get(
        self, id: Optional[Union[int, str]] = None, lazy: bool = False, **kwargs: Any
    ) -> Key:
        if id is not None:
            return cast(Key, super().get(id, lazy=lazy, **kwargs))

        if "fingerprint" not in kwargs:
            raise AttributeError("Missing attribute: id or fingerprint")

        if TYPE_CHECKING:
            assert self.path is not None
        server_data = self.gitlab.http_get(self.path, **kwargs)
        if TYPE_CHECKING:
            assert isinstance(server_data, dict)
        return self._obj_cls(self, server_data)



from typing import Any, cast

from gitlab.base import RESTManager, RESTObject
from gitlab.mixins import GetWithoutIdMixin, RefreshMixin
from gitlab.types import ArrayAttribute

__all__ = [
    "GroupIssuesStatistics",
    "GroupIssuesStatisticsManager",
    "ProjectAdditionalStatistics",
    "ProjectAdditionalStatisticsManager",
    "IssuesStatistics",
    "IssuesStatisticsManager",
    "ProjectIssuesStatistics",
    "ProjectIssuesStatisticsManager",
    "ApplicationStatistics",
    "ApplicationStatisticsManager",
]


class ProjectAdditionalStatistics(RefreshMixin, RESTObject):
    _id_attr = None


class ProjectAdditionalStatisticsManager(GetWithoutIdMixin, RESTManager):
    _path = "/projects/{project_id}/statistics"
    _obj_cls = ProjectAdditionalStatistics
    _from_parent_attrs = {"project_id": "id"}

    def get(self, **kwargs: Any) -> ProjectAdditionalStatistics:
        return cast(ProjectAdditionalStatistics, super().get(**kwargs))


class IssuesStatistics(RefreshMixin, RESTObject):
    _id_attr = None


class IssuesStatisticsManager(GetWithoutIdMixin, RESTManager):
    _path = "/issues_statistics"
    _obj_cls = IssuesStatistics
    _list_filters = ("iids",)
    _types = {"iids": ArrayAttribute}

    def get(self, **kwargs: Any) -> IssuesStatistics:
        return cast(IssuesStatistics, super().get(**kwargs))


class GroupIssuesStatistics(RefreshMixin, RESTObject):
    _id_attr = None


class GroupIssuesStatisticsManager(GetWithoutIdMixin, RESTManager):
    _path = "/groups/{group_id}/issues_statistics"
    _obj_cls = GroupIssuesStatistics
    _from_parent_attrs = {"group_id": "id"}
    _list_filters = ("iids",)
    _types = {"iids": ArrayAttribute}

    def get(self, **kwargs: Any) -> GroupIssuesStatistics:
        return cast(GroupIssuesStatistics, super().get(**kwargs))


class ProjectIssuesStatistics(RefreshMixin, RESTObject):
    _id_attr = None


class ProjectIssuesStatisticsManager(GetWithoutIdMixin, RESTManager):
    _path = "/projects/{project_id}/issues_statistics"
    _obj_cls = ProjectIssuesStatistics
    _from_parent_attrs = {"project_id": "id"}
    _list_filters = ("iids",)
    _types = {"iids": ArrayAttribute}

    def get(self, **kwargs: Any) -> ProjectIssuesStatistics:
        return cast(ProjectIssuesStatistics, super().get(**kwargs))


class ApplicationStatistics(RESTObject):
    _id_attr = None


class ApplicationStatisticsManager(GetWithoutIdMixin, RESTManager):
    _path = "/application/statistics"
    _obj_cls = ApplicationStatistics

    def get(self, **kwargs: Any) -> ApplicationStatistics:
        return cast(ApplicationStatistics, super().get(**kwargs))



from typing import Any, cast, Dict, Optional, Union

from gitlab import exceptions as exc
from gitlab.base import RESTManager, RESTObject
from gitlab.mixins import GetWithoutIdMixin, SaveMixin, UpdateMixin
from gitlab.types import RequiredOptional

__all__ = [
    "ApplicationAppearance",
    "ApplicationAppearanceManager",
]


class ApplicationAppearance(SaveMixin, RESTObject):
    _id_attr = None


class ApplicationAppearanceManager(GetWithoutIdMixin, UpdateMixin, RESTManager):
    _path = "/application/appearance"
    _obj_cls = ApplicationAppearance
    _update_attrs = RequiredOptional(
        optional=(
            "title",
            "description",
            "logo",
            "header_logo",
            "favicon",
            "new_project_guidelines",
            "header_message",
            "footer_message",
            "message_background_color",
            "message_font_color",
            "email_header_and_footer_enabled",
        ),
    )

    @exc.on_http_error(exc.GitlabUpdateError)
    def update(
        self,
        id: Optional[Union[str, int]] = None,
        new_data: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        """Update an object on the server.

        Args:
            id: ID of the object to update (can be None if not required)
            new_data: the update data for the object
            **kwargs: Extra options to send to the server (e.g. sudo)

        Returns:
            The new object data (*not* a RESTObject)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabUpdateError: If the server cannot perform the request
        """
        new_data = new_data or {}
        data = new_data.copy()
        return super().update(id, data, **kwargs)

    def get(self, **kwargs: Any) -> ApplicationAppearance:
        return cast(ApplicationAppearance, super().get(**kwargs))



import time

import pytest

import gitlab


@pytest.fixture
def lazy_project(gl, project):
    assert "/" in project.path_with_namespace
    return gl.projects.get(project.path_with_namespace, lazy=True)


def test_lazy_id(project, lazy_project):
    assert isinstance(lazy_project.id, str)
    assert isinstance(lazy_project.id, gitlab.utils.EncodedId)
    assert lazy_project.id == gitlab.utils.EncodedId(project.path_with_namespace)


def test_refresh_after_lazy_get_with_path(project, lazy_project):
    lazy_project.refresh()
    assert lazy_project.id == project.id


def test_save_after_lazy_get_with_path(project, lazy_project):
    lazy_project.description = "A new description"
    lazy_project.save()
    assert lazy_project.id == project.id
    assert lazy_project.description == "A new description"


def test_delete_after_lazy_get_with_path(gl, group):
    project = gl.projects.create({"name": "lazy_project", "namespace_id": group.id})
    # Pause to let GL catch up (happens on hosted too, sometimes takes a while for server to be ready to merge)
    time.sleep(5)
    lazy_project = gl.projects.get(project.path_with_namespace, lazy=True)
    lazy_project.delete()


def test_list_children_after_lazy_get_with_path(gl, lazy_project):
    lazy_project.mergerequests.list()



"""
GitLab API:
https://docs.gitlab.com/ce/api/merge_requests.html
https://docs.gitlab.com/ee/api/deployments.html#list-of-merge-requests-associated-with-a-deployment
"""

import re

import pytest
import responses

from gitlab.v4.objects import (
    ProjectDeploymentMergeRequest,
    ProjectMergeRequest,
    ProjectMergeRequestReviewerDetail,
)

mr_content = {
    "id": 1,
    "iid": 1,
    "project_id": 3,
    "title": "test1",
    "description": "fixed login page css paddings",
    "state": "merged",
    "merged_by": {
        "id": 87854,
        "name": "Douwe Maan",
        "username": "DouweM",
        "state": "active",
        "avatar_url": "https://gitlab.example.com/uploads/-/system/user/avatar/87854/avatar.png",
        "web_url": "https://gitlab.com/DouweM",
    },
    "reviewers": [
        {
            "id": 2,
            "name": "Sam Bauch",
            "username": "kenyatta_oconnell",
            "state": "active",
            "avatar_url": "https://www.gravatar.com/avatar/956c92487c6f6f7616b536927e22c9a0?s=80&d=identicon",
            "web_url": "http://gitlab.example.com//kenyatta_oconnell",
        }
    ],
}

reviewers_content = [
    {
        "user": {
            "id": 2,
            "name": "Sam Bauch",
            "username": "kenyatta_oconnell",
            "state": "active",
            "avatar_url": "https://www.gravatar.com/avatar/956c92487c6f6f7616b536927e22c9a0?s=80&d=identicon",
            "web_url": "http://gitlab.example.com//kenyatta_oconnell",
        },
        "state": "unreviewed",
        "created_at": "2022-07-27T17:03:27.684Z",
    }
]


@pytest.fixture
def resp_list_merge_requests():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url=re.compile(
                r"http://localhost/api/v4/projects/1/(deployments/1/)?merge_requests"
            ),
            json=[mr_content],
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_get_merge_request_reviewers():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/projects/1/merge_requests/1",
            json=mr_content,
            content_type="application/json",
            status=200,
        )
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/projects/3/merge_requests/1/reviewers",
            json=reviewers_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


def test_list_project_merge_requests(project, resp_list_merge_requests):
    mrs = project.mergerequests.list()
    assert isinstance(mrs[0], ProjectMergeRequest)
    assert mrs[0].iid == mr_content["iid"]


def test_list_deployment_merge_requests(project, resp_list_merge_requests):
    deployment = project.deployments.get(1, lazy=True)
    mrs = deployment.mergerequests.list()
    assert isinstance(mrs[0], ProjectDeploymentMergeRequest)
    assert mrs[0].iid == mr_content["iid"]


def test_get_merge_request_reviewers(project, resp_get_merge_request_reviewers):
    mr = project.mergerequests.get(1)
    reviewers_details = mr.reviewer_details.list()
    assert isinstance(mr, ProjectMergeRequest)
    assert isinstance(reviewers_details, list)
    assert isinstance(reviewers_details[0], ProjectMergeRequestReviewerDetail)
    assert mr.reviewers[0]["name"] == reviewers_details[0].user["name"]
    assert reviewers_details[0].state == "unreviewed"
    assert reviewers_details[0].created_at == "2022-07-27T17:03:27.684Z"
